import numpy as np
import pytest
import torch

from fibber.datasets.dataset_utils import get_demo_dataset
from fibber.metrics.bert_classifier import BertClassifier
from fibber.resources.resource_utils import get_bert_clf_demo


@pytest.fixture()
def gpu_id():
    if torch.cuda.device_count() > 0:
        return 0
    return -1


@pytest.fixture()
def bert_classifier_on_demo(gpu_id):
    get_bert_clf_demo()
    trainset, testset = get_demo_dataset()
    bert_classifier = BertClassifier(
        "demo", trainset, testset, bert_gpu_id=gpu_id, bert_clf_steps=5000)
    return bert_classifier


@pytest.mark.slow
def test_bert_classifier(bert_classifier_on_demo):
    assert isinstance(bert_classifier_on_demo, BertClassifier)

    io_pairs = [
        ("This is a bad movie", 0),
        ("This is a good movie", 1)
    ]

    for x, y in io_pairs:
        dist = bert_classifier_on_demo.predict_dist_example(None, x)
        assert len(dist) == 2
        assert np.argmax(dist) == y

        pred = bert_classifier_on_demo.predict_example(None, x)
        assert pred == y

        pred = bert_classifier_on_demo.measure_example(None, x)
        assert pred == y

    batched_io_pairs = [
        (["This is a bad movie.",
          "This is a good movie. I want to buy a ticket."],
         [0, 1])
    ]

    for x, y in batched_io_pairs:
        dist = bert_classifier_on_demo.predict_dist_batch(None, x)
        assert dist.shape[0] == 2 and dist.shape[1] == 2
        assert np.all(np.argmax(dist, axis=1) == y)

        pred = bert_classifier_on_demo.predict_batch(None, x)
        assert all([a == b for a, b in zip(pred, y)])

        pred = bert_classifier_on_demo.measure_batch(None, x)
        assert all([a == b for a, b in zip(pred, y)])



"""This module provides utility functions and classes to handle fibber's datasets.

* To load a dataset, use ``get_dataset`` function. For example, to load AG's news dataset, run::

    trainset, testset =  get_dataset("ag")

* The trainset and testset are both dicts. The dict looks like::

    {
      "label_mapping": [
        "World",
        "Sports",
        "Business",
        "Sci/Tech"
      ],
      "cased": true,
      "paraphrase_field": "text0",
      "data": [
        {
          "label": 1,
          "text0": "Boston won the NBA championship in 2008."
        },
        {
          "label": 3,
          "text0": "Apple releases its latest cell phone."
        },
        ...
      ]
    }

* To sub-sample 100 examples from training set, run::

    subsampled_dataset = subsample_dataset(trainset, 100)

* To convert a dataset dict to a ``torch.IterableDataset`` for BERT model, run::

    iterable_dataset = DatasetForBert(trainset, "bert-base-cased", batch_size=32);

For more details, see ``https://dai-lab.github.io/fibber/``
"""

import copy
import hashlib
import json
import os

import numpy as np
import torch
import tqdm
from transformers import BertTokenizerFast

from fibber import get_root_dir, log, resources
from fibber.datasets.downloadable_datasets import downloadable_dataset_urls
from fibber.download_utils import download_file

logger = log.setup_custom_logger(__name__)


def get_dataset(dataset_name):
    """Load dataset from fibber root directory.

    Users should make sure the data is downloaded to the ``datasets`` folder in fibber root
    dir (default: ``~/.fibber/datasets``). Otherwise, assertion error is raised.

    Args:
        dataset_name (str): the name of the dataset. See ``https://dai-lab.github.io/fibber/``
            for a full list of built-in datasets.

    Returns:
        (dict, dict): the function returns a tuple of two dict, representing the training set and
        test set respectively.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "datasets")

    if dataset_name == "mnli" or dataset_name == "mnli_mis":
        train_filename = os.path.join(data_dir, "mnli/train.json")
        if dataset_name == "mnli":
            test_filename = os.path.join(data_dir, "mnli/dev_matched.json")
        else:
            test_filename = os.path.join(data_dir, "mnli/dev_mismatched.json")

    else:
        train_filename = os.path.join(data_dir, dataset_name, "train.json")
        test_filename = os.path.join(data_dir, dataset_name, "test.json")

    if not os.path.exists(train_filename) or not os.path.exists(test_filename):
        logger.error("%s dataset not found.", dataset_name)
        assert 0, ("Please use `python3 -m fibber.datasets.download_datasets` "
                   "to download datasets.")

    with open(train_filename) as f:
        trainset = json.load(f)

    with open(test_filename) as f:
        testset = json.load(f)

    logger.info("%s training set has %d records.", dataset_name, len(trainset["data"]))
    logger.info("%s test set has %d records.", dataset_name, len(testset["data"]))
    return trainset, testset


def get_demo_dataset():
    """download demo dataset.

    Returns:
        (dict, dict): trainset and testset.
    """
    download_file(subdir="", **downloadable_dataset_urls["mr-demo"])

    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "mr-demo")

    with open(os.path.join(data_dir, "train.json")) as f:
        trainset = json.load(f)
    with open(os.path.join(data_dir, "test.json")) as f:
        testset = json.load(f)

    logger.info("Demo training set has %d records.", len(trainset["data"]))
    logger.info("Demo test set has %d records.", len(testset["data"]))

    return trainset, testset


def text_md5(x):
    """Computes and returns the md5 hash of a str."""
    m = hashlib.md5()
    m.update(x.encode('utf8'))
    return m.hexdigest()


def subsample_dataset(dataset, n):
    """Sub-sample a dataset to `n` examples.

    Data is selected evenly and randomly from each category. Data in each category is sorted by
    its md5 hash value. The top ``(n // k)`` examples from each category are included in the
    sub-sampled dataset, where ``k`` is the number of categories.

    If ``n`` is not divisible by ``k``, one more data is sampled from the first ``(n % k)``
    categories.

    If the dataset has less than ``n`` examples, a copy of the original dataset will be returned.

    Args:
        dataset (dict): a dataset dict.
        n (int): the size of the sub-sampled dataset.

    Returns:
        (dict): a sub-sampled dataset as a dict.
    """
    if n > len(dataset["data"]):
        return copy.deepcopy(dataset)

    bins = [[] for i in dataset["label_mapping"]]

    subset = dict([(k, v) for k, v in dataset.items() if k != "data"])

    for idx, data in enumerate(dataset["data"]):
        label = data["label"]
        text = data["text0"]
        if "text1" in data:
            text += data["text1"]

        bins[label].append((idx, text_md5(text)))

    data_list = []
    for i in range(len(bins)):
        bins[i] = sorted(bins[i], key=lambda x: x[1])
        m = n // len(bins) + (1 if i < n % len(bins) else 0)
        for j in range(m):
            data_list.append(copy.deepcopy(dataset["data"][bins[i][j][0]]))

    subset["data"] = data_list

    return subset


def verify_dataset(dataset):
    """Verify if the dataset dict contains necessary fields.

    Assertion error is raised if there are missing or incorrect fields.

    Args:
        dataset (dict): a dataset dict.
    """
    assert "label_mapping" in dataset
    assert "cased" in dataset
    assert "paraphrase_field" in dataset
    assert dataset["paraphrase_field"] in ["text0", "text1"]

    num_labels = len(dataset["label_mapping"])
    counter = [0] * num_labels

    for data_record in tqdm.tqdm(dataset["data"]):
        assert "label" in data_record
        label = data_record["label"]
        assert 0 <= label < num_labels
        counter[label] += 1
        assert "text0" in data_record
        assert dataset["paraphrase_field"] in data_record

    for item in counter:
        assert item > 0, "empty class"


class DatasetForBert(torch.utils.data.IterableDataset):
    """Create a ``torch.IterableDataset`` for a BERT model.

    The module is an iterator that yields infinite batches from the dataset. To construct a
    batch, we randomly sample a few examples with similar length. Then we pad all selected
    examples to the same length ``L``. Then we construct a tuple of 4 or 5 tensors. All
    tensors are on CPU.

    Each example starts with ``[CLS]``, and ends with ``[SEP]``. If there are two parts in
    the input, the two parts are separated by ``[SEP]``.

    __iter__(self):

    Yields:
        A tuple of tensors.

        * The first tensor is an int tensor of size ``(batch_size, L)``, representing
          word ids. Each row of this tensor correspond to one example in the dataset.
          If ``masked_lm == True``, the tensor stores the masked text.
        * The second tensor is an int tensor of size ``(batch_size, L)``, representing
          the text length. Each entry is 1 if the corresponding position is text, and it
          is 0 if the position is padding.
        * The third tensor is an int tensor of size ``(batch_size, L)``, representing the
          token type. The token type is 0 if current position is in the first part of the
          input text. And it is 1 if current position is in the second part of the input.
          For padding positions, token type is 0.
        * The forth tensor an int tensor of size ``(batch_size,)``, representing the
          classification label.
        * (optional) If ``masked_lm == True``, the fifth tensor is a tensor of size
          ``(batch_size, L)``. Each entry in this tensor is either -100 if the position is
          not masked, or the correct word if the position is masked. Note that, a masked
          position is not always a ``[MASK]`` token in the first tensor. With 80%
          probability, it is a ``[MASK]``. With 10% probability, it is the original word.
          And with 10% probability, it is a random word.
    """

    def __init__(self, dataset, model_init, batch_size, exclude=-1,
                 masked_lm=False, masked_lm_ratio=0.2, seed=0):
        """Initialize.

        Args:
            dataset (dict): a dataset dict.
            model_init (str): the pre-trained model name. select from ``['bert-base-cased',
                'bert-base-uncased', 'bert-large-cased', and 'bert-large-uncased']``.
            batch_size (int): the batch size in each step.
            exclude (int): exclude one category from the data.
                Use -1 (default) to include all categories.
            masked_lm (bool): whether to randomly replace words with mask tokens.
            masked_lm_ratio (float): the ratio of random masks. Ignored when masked_lm is False.
            seed: random seed.
        """
        self._buckets = [30, 50, 100, 200]
        self._max_len = self._buckets[-1]
        self._data = [[] for i in range(len(self._buckets))]

        self._batch_size = batch_size
        self._tokenizer = BertTokenizerFast.from_pretrained(
            resources.get_transformers(model_init), do_lower_case="uncased" in model_init)

        self._seed = seed
        self._pad_tok_id = self._tokenizer.pad_token_id

        self._masked_lm = masked_lm
        self._masked_lm_ratio = masked_lm_ratio
        self._mask_tok_id = self._tokenizer.mask_token_id

        counter = 0
        logger.info("DatasetForBert is processing data.")
        for item in tqdm.tqdm(dataset["data"]):
            y = item["label"]
            s0 = "[CLS] " + item["text0"] + " [SEP]"
            if "text1" in item:
                s1 = item["text1"] + " [SEP]"
            else:
                s1 = ""

            if y == exclude:
                continue

            counter += 1

            s0_ids = self._tokenizer.convert_tokens_to_ids(
                self._tokenizer.tokenize(s0))
            s1_ids = self._tokenizer.convert_tokens_to_ids(
                self._tokenizer.tokenize(s1))
            text_ids = (s0_ids + s1_ids)[:self._max_len]

            for bucket_id in range(len(self._buckets)):
                if self._buckets[bucket_id] >= len(text_ids):
                    self._data[bucket_id].append((text_ids, y, len(s0_ids), len(s1_ids)))
                    break

        logger.info("Load %d documents. with filter %d.", counter, exclude)
        self._bucket_prob = np.asarray([len(x) for x in self._data])
        self._bucket_prob = self._bucket_prob / np.sum(self._bucket_prob)

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        if worker_info is None:
            self._rng = np.random.RandomState(self._seed)
        else:
            self._rng = np.random.RandomState(worker_info.id + self._seed)

        while True:
            bucket_id = self._rng.choice(len(self._bucket_prob), p=self._bucket_prob)
            ids = self._rng.choice(len(self._data[bucket_id]), self._batch_size)

            texts, labels, l0s, l1s = zip(*[self._data[bucket_id][id] for id in ids])

            text_len = [len(x) for x in texts]
            max_text_len = max(text_len)

            # tok_list is copied by list add.
            texts = [x + [self._pad_tok_id] * (max_text_len - len(x)) for x in texts]
            masks = [[1] * x + [0] * (max_text_len - x) for x in text_len]
            tok_types = [([0] * l0 + [1] * l1 + [0] * (max_text_len - l0 - l1)
                          )[:max_text_len]
                         for (l0, l1) in zip(l0s, l1s)]

            texts = np.asarray(texts)
            masks = np.asarray(masks)
            labels = np.asarray(labels)
            tok_types = np.asarray(tok_types)

            if not self._masked_lm:
                yield (torch.tensor(texts), torch.tensor(masks),
                       torch.tensor(tok_types), torch.tensor(labels))
            else:
                rand_t = self._rng.rand(self._batch_size, max_text_len)
                masked_pos = (rand_t < self._masked_lm_ratio) * masks
                masked_pos[:, 0] = 0
                lm_labels = (masked_pos * texts - 100 * (1 - masked_pos))
                rand_t = self._rng.rand(self._batch_size, max_text_len)
                filling = self._rng.randint(0, self._tokenizer.vocab_size,
                                            (self._batch_size, max_text_len))
                filling = ((rand_t < 0.8) * self._mask_tok_id
                           + (rand_t >= 0.8) * (rand_t < 0.9) * filling
                           + (rand_t >= 0.9) * texts)
                texts = masked_pos * filling + (1 - masked_pos) * texts
                yield (torch.tensor(texts), torch.tensor(masks), torch.tensor(tok_types),
                       torch.tensor(labels), torch.tensor(lm_labels))



import json

import tqdm

from fibber import log
from fibber.datasets.preprocess_utils import download_raw_and_preprocess

logger = log.setup_custom_logger(__name__)


def preprocess_mnli_data(input_filename, output_filename):
    """preprocess raw MNLI dataset to Fibber's JSON format."""
    logger.info("Start preprocessing data, and save at %s.", output_filename)

    with open(input_filename) as f:
        df = f.readlines()
    df = [json.loads(line) for line in df]

    mapping = {
        "neutral": 0,
        "entailment": 1,
        "contradiction": 2
    }
    data = {
        "label_mapping": ["neutral", "entailment", "contradiction"],
        "cased": True,
        "paraphrase_field": "text1",
    }

    datalist = []
    for item in tqdm.tqdm(df):
        if item["gold_label"] == '-':
            continue
        y = mapping[item["gold_label"]]
        datalist.append({
            "label": y,
            "text0": item["sentence1"],
            "text1": item["sentence2"]
        })

    data["data"] = datalist
    with open(output_filename, "w") as f:
        json.dump(data, f, indent=2)


def download_and_preprocess_mnli():
    """Download and preprocess MNLI dataset."""
    download_raw_and_preprocess(
        dataset_name="mnli",
        download_list=["mnli-raw"],
        preprocess_fn=preprocess_mnli_data,
        preprocess_input_output_list=[
            ("raw/multinli_1.0/multinli_1.0_train.jsonl", "train.json"),
            ("raw/multinli_1.0/multinli_1.0_dev_matched.jsonl", "dev_matched.json"),
            ("raw/multinli_1.0/multinli_1.0_dev_mismatched.jsonl", "dev_mismatched.json")])


if __name__ == "__main__":
    download_and_preprocess_mnli()



downloadable_dataset_urls = {
    "ag-raw-train": {
        "filename": "train.csv",
        "url": ("https://raw.githubusercontent.com/mhjabreel/CharCnn_Keras/master/"
                "data/ag_news_csv/train.csv"),
        "md5": "b1a00f826fdfbd249f79597b59e1dc12",
        "untar": False,
        "unzip": False,
    },
    "ag-raw-test": {
        "filename": "test.csv",
        "url": ("https://raw.githubusercontent.com/mhjabreel/CharCnn_Keras/master/data/"
                "ag_news_csv/test.csv"),
        "md5": "d52ea96a97a2d943681189a97654912d",
        "untar": False,
        "unzip": False,
    },
    "imdb-raw": {
        "filename": "aclImdb_v1.tar.gz",
        "url": "https://ai.stanford.edu/~amaas/data/sentiment/aclImdb_v1.tar.gz",
        "md5": "7c2ac02c03563afcf9b574c7e56c153a",
        "untar": True,
        "unzip": False,
    },
    "yelp-raw": {
        "filename": "yelp_review_polarity_csv.tgz",
        "url": "https://s3.amazonaws.com/fast-ai-nlp/yelp_review_polarity_csv.tgz",
        "md5": "0f09b3af1a79c136ef9ca5f29df9ed9a",
        "untar": True,
        "unzip": False,
    },
    "mr-raw": {
        "filename": "rt-polaritydata.tar.gz",
        "url": "http://www.cs.cornell.edu/people/pabo/movie-review-data/rt-polaritydata.tar.gz",
        "md5": "50c1c2c047b4225e148e97aa7708c34e",
        "untar": True,
        "unzip": False,
    },
    "snli-raw": {
        "filename": "snli_1.0.zip",
        "url": "https://nlp.stanford.edu/projects/snli/snli_1.0.zip",
        "md5": "981c3df556bbaea3f17f752456d0088c",
        "untar": False,
        "unzip": True,
    },
    "mnli-raw": {
        "filename": "multinli_1.0.zip",
        "url": "https://cims.nyu.edu/~sbowman/multinli/multinli_1.0.zip",
        "md5": "0f70aaf66293b3c088a864891db51353",
        "untar": False,
        "unzip": True,
    },
    "processed-datasets": {
        "filename": "datasets.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/datasets_v0.2.tgz",
        "md5": "81b1eda8421764a0b648cb83425f53a3",
        "untar": True,
        "unzip": False,
    },
    "mr-demo": {
        "filename": "mr-demo.tgz",
        "url": "https://fibber-data.s3.amazonaws.com/mr-demo.tgz",
        "md5": "8ee6b6eda12a7c1282cb903c713085c5",
        "untar": True,
        "unzip": False,
    }
}



import os

import numpy as np
import torch
import tqdm
from nltk import word_tokenize
from torch import nn
from transformers import BertTokenizer

from fibber import log, resources
from fibber.resources import get_glove_emb, get_stopwords

logger = log.setup_custom_logger(__name__)


class WordPieceDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, model_init):
        self._tokenizer = BertTokenizer.from_pretrained(
            resources.get_transformers(model_init), do_lower_case="uncased" in model_init)

        self._glove = get_glove_emb()
        stopwords = get_stopwords()

        for word in stopwords:
            word = word.lower().strip()
            if word in self._glove["tok2id"]:
                self._glove["emb_table"][self._glove["tok2id"][word], :] = 0

        data = []
        logger.info("processing data for wordpiece embedding training")
        for item in tqdm.tqdm(dataset["data"]):
            text = item["text0"]
            if "text1" in item:
                text += " " + item["text1"]

            text_toks = word_tokenize(text)
            data += [x for x in text_toks if x.lower() in self._glove["tok2id"]]
        self._data = data

    def __len__(self):
        return len(self._data)

    def __getitem__(self, i):
        w = self._data[i]

        g_emb = self._glove["emb_table"][self._glove["tok2id"][w.lower()]]

        comp = np.zeros(self._tokenizer.vocab_size)
        for tok_id in self._tokenizer.convert_tokens_to_ids(
                self._tokenizer.tokenize(w)):
            comp[tok_id] = 1

        return torch.tensor(comp).float(), torch.tensor(g_emb).float()


def get_wordpiece_emb(output_dir, dataset_name, trainset, device,
                      steps=500, bs=1000, lr=1, lr_halve_steps=100):
    """Transfer GloVe embeddings to BERT vocabulary.

    The transfered embeddings will be stored at ``<output_dir>/wordpiece_emb*``.

    Args:
        output_dir (str): a directory to store pretrained language model.
        dataset_name (str): dataset name.
        trainset (dict): the dataset dist.
        device (torch.Device): a device to train the model.
        steps (int): transfering steps.
        bs (int): transfering batch size.
        lr (str): transfering learning rate.
        lr_halve_steps (int): steps to halve the learning rate.
    Returns:
        (np.array): a array of size (300, N) where N is the vocabulary size for a bert-base model.
    """
    filename = output_dir + "/wordpiece_emb-%s-%04d.pt" % (
        dataset_name, steps)
    if os.path.exists(filename):
        state_dict = torch.load(filename)
        logger.info("load wordpiece embeddings from %s", filename)
        return state_dict["embs"]

    if trainset["cased"]:
        model_init = "bert-base-cased"
    else:
        model_init = "bert-base-uncased"

    dataset = WordPieceDataset(trainset, model_init)
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=bs, shuffle=True,
        num_workers=0, drop_last=True)

    linear = nn.Linear(dataset._tokenizer.vocab_size, 300, bias=False).to(device)
    opt = torch.optim.SGD(lr=lr, momentum=0.5,
                          params=linear.parameters())
    scheduler = torch.optim.lr_scheduler.StepLR(
        opt, step_size=lr_halve_steps, gamma=0.5)

    for w, wid in dataset._tokenizer.vocab.items():
        if w.lower() in dataset._glove["tok2id"]:
            linear.weight.data[:, wid] = torch.tensor(
                dataset._glove["emb_table"][dataset._glove["tok2id"][w.lower()]]).to(device)

    logger.info("train word piece embeddings")
    pbar = tqdm.tqdm(total=steps)
    losses = []
    global_step = 0
    while True:
        for idx, (comp, target) in enumerate(dataloader):
            comp = comp.to(device)
            target = target.to(device)
            pred_emb = linear(comp)
            loss = ((pred_emb - target).abs()).sum(dim=1).mean(dim=0)
            losses.append(loss.detach().cpu().numpy())
            opt.zero_grad()
            loss.backward()
            opt.step()
            scheduler.step()
            losses = losses[-20:]

            pbar.set_postfix(loss=np.mean(losses), std=target.std().detach().cpu().numpy())
            pbar.update(1)
            global_step += 1
            if global_step >= steps:
                break
        if global_step >= steps:
            break

    pbar.close()
    wordpiece_emb = linear.weight.data.cpu().numpy()
    torch.save({"embs": wordpiece_emb}, filename)
    return wordpiece_emb



import argparse
import datetime
import os

from fibber import log
from fibber.benchmark.benchmark_utils import update_detailed_result
from fibber.datasets import builtin_datasets, get_dataset, subsample_dataset, verify_dataset
from fibber.metrics.attack_aggregation_utils import add_sentence_level_adversarial_attack_metrics
from fibber.metrics.metric_utils import MetricBundle
from fibber.paraphrase_strategies import (
    BertSamplingStrategy, IdentityStrategy, RandomStrategy, TextAttackStrategy)
from fibber.paraphrase_strategies.strategy_base import StrategyBase

logger = log.setup_custom_logger(__name__)
log.remove_logger_tf_handler(logger)

built_in_strategies = {
    "RandomStrategy": RandomStrategy,
    "IdentityStrategy": IdentityStrategy,
    "TextAttackStrategy": TextAttackStrategy,
    "BertSamplingStrategy": BertSamplingStrategy
}

DATASET_NAME_COL = "0_dataset_name"
STRATEGY_NAME_COL = "1_paraphrase_strategy_name"


class Benchmark(object):
    """Benchmark framework for adversarial attack methods on text classification."""

    def __init__(self,
                 output_dir, dataset_name,
                 trainset=None, testset=None, attack_set=None,
                 subsample_attack_set=0,
                 customized_clf=None,
                 enable_bert_clf=True,
                 use_gpu_id=-1,
                 gpt2_gpu_id=-1,
                 bert_gpu_id=-1,
                 bert_clf_steps=20000,
                 bert_clf_bs=32):
        """Initialize Benchmark framework.

        Args:
            output_dir (str): the directory to write outputs including model, sentences, metrics
                and log.
            dataset_name (str): the name of the dataset.
            trainset (dict): the training set. If the ``dataset_name`` matches built-in datasets,
                ``trainset`` should be None.
            testset (dict): the test set. If the ``dataset_name`` matches built-in datasets,
                ``testset`` should be None.
            attack_set (dict or None): the set to run adversarial attack. Use None to attack the
                ``testset``.
            subsample_attack_set (int): subsample the attack set. 0 to use the whole attack set.
            customized_clf (MetricBase): an classifier object.
            enable_bert_clf (bool): whether to enable bert classifier in metrics. You can disable
                it when you are attacking your own classifier.
            use_gpu_id (int): the gpu to run universal sentence encoder to compute metrics.
                -1 for CPU.
            gpt2_gpu_id (int): the gpu to run the GPT2-medium language model to compute metrics.
                -1 for CPU.
            bert_gpu_id (int): the gpu to run the BERT text classifier, which is the model being
                attacked. -1 for CPU.
            bert_clf_steps (int): number of steps to train the BERT text classifier.
            bert_clf_bs (int): the batch size to train the BERT classifier.
        """
        # make output dir
        self._output_dir = output_dir
        os.makedirs(output_dir, exist_ok=True)
        self._dataset_name = dataset_name

        # setup dataset
        if dataset_name in builtin_datasets:
            if trainset is not None or testset is not None:
                logger.error(("dataset name %d conflict with builtin dataset. "
                              "set trainset and testset to None.") % dataset_name)
                raise RuntimeError
            trainset, testset = get_dataset(dataset_name)
        else:
            verify_dataset(trainset)
            verify_dataset(testset)

        if attack_set is None:
            attack_set = testset

        if subsample_attack_set != 0:
            attack_set = subsample_dataset(attack_set, subsample_attack_set)

        self._trainset = trainset
        self._testset = testset
        self._attack_set = attack_set

        # setup metric bundle
        self._metric_bundle = MetricBundle(
            enable_bert_clf_prediction=enable_bert_clf,
            use_gpu_id=use_gpu_id, gpt2_gpu_id=gpt2_gpu_id,
            bert_gpu_id=bert_gpu_id, dataset_name=dataset_name,
            trainset=self._trainset, testset=testset,
            bert_clf_steps=bert_clf_steps,
            bert_clf_bs=bert_clf_bs
        )

        if customized_clf:
            self._metric_bundle.add_classifier(str(customized_clf), customized_clf)
            self._metric_bundle.set_target_classifier_by_name(str(customized_clf))

        add_sentence_level_adversarial_attack_metrics(
            self._metric_bundle, gpt2_ppl_threshold=5, use_sim_threshold=0.85)

    def run_benchmark(self,
                      paraphrase_strategy="IdentityStrategy",
                      strategy_gpu_id=-1,
                      num_paraphrases_per_text=50,
                      exp_name=None,
                      update_global_results=True):
        """Run the benchmark.

        Args:
            paraphrase_strategy (str or StrategyBase): the paraphrase strategy to benchmark.
                Either the name of a builtin strategy or a customized strategy derived from
                StrategyBase.
            strategy_gpu_id (int): the gpu id to run the strategy. -1 for CPU. Ignored when
                ``paraphrase_strategy`` is an object.
            num_paraphrases_per_text (int): number of paraphrases for each sentence.
            exp_name (str or None): the name of current experiment. None for default name. the
                default name is ``<dataset_name>-<strategy_name>-<date>-<time>``.
            update_global_results (bool): whether to write results in <fibber_root_dir> or the
                benchmark output dir.

        Returns:
            A dict of evaluation results.
        """
        # setup strategy
        if isinstance(paraphrase_strategy, str):
            if paraphrase_strategy in built_in_strategies:
                paraphrase_strategy = built_in_strategies[paraphrase_strategy](
                    {}, self._dataset_name, strategy_gpu_id, self._output_dir, self._metric_bundle)
        else:
            assert isinstance(paraphrase_strategy, StrategyBase)

        # get experiment name
        if exp_name is None:
            exp_name = (self._dataset_name + "-" + str(paraphrase_strategy) + "-"
                        + datetime.datetime.now().strftime("%m%d-%H%M%S"))

        paraphrase_strategy.fit(self._trainset)
        tmp_output_filename = os.path.join(
            self._output_dir, exp_name + "-tmp.json")
        logger.info("Write paraphrase temporary results in %s.", tmp_output_filename)
        results = paraphrase_strategy.paraphrase_dataset(
            self._attack_set, num_paraphrases_per_text, tmp_output_filename)

        output_filename = os.path.join(
            self._output_dir, exp_name + "-with-metric.json")
        logger.info("Write paraphrase with metrics in %s.", tmp_output_filename)

        results = self._metric_bundle.measure_dataset(
            results=results, output_filename=output_filename)

        aggregated_result = self._metric_bundle.aggregate_metrics(
            self._dataset_name, str(paraphrase_strategy), exp_name, results)
        update_detailed_result(aggregated_result,
                               self._output_dir if not update_global_results else None)
        return aggregated_result

    def get_metric_bundle(self):
        return self._metric_bundle


def get_strategy(arg_dict, dataset_name, strategy_name, strategy_gpu_id,
                 output_dir, metric_bundle):
    """Take the strategy name and construct a strategy object."""
    if strategy_name == "RandomStrategy":
        return RandomStrategy(arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "IdentityStrategy":
        return IdentityStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "TextAttackStrategy":
        return TextAttackStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    if strategy_name == "BertSamplingStrategy":
        return BertSamplingStrategy(
            arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle)
    else:
        assert 0


def main():
    parser = argparse.ArgumentParser()

    # add experiment args
    parser.add_argument("--dataset", type=str, default="ag")
    parser.add_argument("--output_dir", type=str, default=None)
    parser.add_argument("--num_paraphrases_per_text", type=int, default=20)
    parser.add_argument("--subsample_testset", type=int, default=1000)
    parser.add_argument("--strategy", type=str, default="RandomStrategy")
    parser.add_argument("--strategy_gpu_id", type=int, default=-1)

    # metric args
    parser.add_argument("--gpt2_gpu_id", type=int, default=-1)
    parser.add_argument("--bert_gpu_id", type=int, default=-1)
    parser.add_argument("--use_gpu_id", type=int, default=-1)
    parser.add_argument("--bert_clf_steps", type=int, default=20000)

    # add builtin strategies' args to parser.
    for item in built_in_strategies.values():
        item.add_parser_args(parser)

    arg_dict = vars(parser.parse_args())
    assert arg_dict["output_dir"] is not None

    os.makedirs(arg_dict["output_dir"], exist_ok=True)
    os.makedirs(os.path.join(arg_dict["output_dir"], "log"), exist_ok=True)

    benchmark = Benchmark(arg_dict["output_dir"], arg_dict["dataset"],
                          subsample_attack_set=arg_dict["subsample_testset"],
                          use_gpu_id=arg_dict["use_gpu_id"],
                          bert_gpu_id=arg_dict["bert_gpu_id"],
                          gpt2_gpu_id=arg_dict["gpt2_gpu_id"],
                          bert_clf_steps=arg_dict["bert_clf_steps"])

    log.add_file_handler(
        logger, os.path.join(arg_dict["output_dir"], "log.log"))
    log.remove_logger_tf_handler(logger)

    # Get paraphrase strategy
    paraphrase_strategy = get_strategy(arg_dict, arg_dict["dataset"], arg_dict["strategy"],
                                       arg_dict["strategy_gpu_id"], arg_dict["output_dir"],
                                       benchmark.get_metric_bundle())

    benchmark.run_benchmark(paraphrase_strategy=paraphrase_strategy,
                            num_paraphrases_per_text=arg_dict["num_paraphrases_per_text"])


if __name__ == "__main__":
    main()



"""This module implements the paraphrase strategy using TextFooler."""

import sys

from fibber import log
from fibber.paraphrase_strategies.strategy_base import StrategyBase

logger = log.setup_custom_logger(__name__)

try:
    import textattack
    from textattack import attack_recipes
    from textattack.models.wrappers.model_wrapper import ModelWrapper
except ImportError:
    logger.warning("TextAttack is not installed so TextFooler stategy can't be used. "
                   "Install it by `pip install textattack`.")
    ModelWrapper = object


class CLFModel(ModelWrapper):
    """A classifier wrapper for textattack package."""

    def __init__(self, clf_metric, field_name):
        self.model = clf_metric
        self._field_name = field_name
        self._tokenizer = clf_metric._tokenizer
        self._data_record = None

    def __call__(self, text_list):
        ret = self.model.predict_dist_batch(
            self._data_record[self._field_name], text_list,
            data_record=self._data_record, paraphrase_field=self._field_name)
        return ret

    def set_data_record(self, data_record):
        self._data_record = data_record

    def tokenize(self, inputs):
        """Helper method that calls ``tokenizer.batch_encode`` if possible, and
        if not, falls back to calling ``tokenizer.encode`` for each input."""
        if hasattr(self._tokenizer, "batch_encode"):
            return self._tokenizer.batch_encode(inputs)
        else:
            return [self._tokenizer.encode(x) for x in inputs]


class TextAttackStrategy(StrategyBase):
    """This strategy is a wrapper for strategies implemented in TextAttack Package.

    The recipe is used to attack the classifier in metric_bundle.
        If the attack succeeds, we use the adversarial sentence as a paraphrase.
        If the attack fails, we use the original sentence as a paraphrase.

    This strategy always returns one paraphrase for one data record, regardless of `n`.
    """

    __abbr__ = "ta"
    __hyperparameters__ = [
        ("recipe", str, "TextFoolerJin2019", "an attacking recipe implemented in TextAttack.")
    ]

    def __init__(self, arg_dict, dataset_name, strategy_gpu_id, output_dir, metric_bundle):
        if "textattack" not in sys.modules:
            logger.error("TextAttack not installed. Please install textattack manually.")
            raise RuntimeError
        super(TextAttackStrategy, self).__init__(arg_dict, dataset_name, strategy_gpu_id,
                                                 output_dir, metric_bundle)

    def __repr__(self):
        return self._strategy_config["recipe"]

    def fit(self, trainset):
        if ModelWrapper is None:
            raise RuntimeError("no internet connection.")

        self._model = CLFModel(self._metric_bundle.get_target_classifier(),
                               trainset["paraphrase_field"])
        self._recipe = getattr(attack_recipes, self._strategy_config["recipe"]
                               ).build(self._model)

    def paraphrase_example(self, data_record, field_name, n):
        """Generate paraphrased sentences."""
        self._model.set_data_record(data_record)

        attack_text = " ".join(data_record[field_name].split()[:200])
        att = next(self._recipe.attack_dataset(
            [(attack_text, data_record["label"])]))
        if isinstance(att, textattack.attack_results.SuccessfulAttackResult):
            return [att.perturbed_result.attacked_text.text]
        else:
            return [data_record[field_name]]



import os

import numpy as np
import tqdm

from fibber import log
from fibber.download_utils import download_file, get_root_dir
from fibber.resources.downloadable_resources import downloadable_resource_urls

logger = log.setup_custom_logger(__name__)


def load_glove_model(glove_file, dim):
    """Load glove embeddings from txt file.

    Args:
        glove_file: filename.
        dim: the dimension of the embedding.

    Returns:
        a dict:
            "emb_table": a numpy array of size(N, 300)
            "id2tok": a list of strings.
            "tok2id": a dict that maps word (string) to its id.
    """
    glove_file_lines = open(glove_file, "r", encoding="utf8").readlines()

    emb_table = np.zeros((len(glove_file_lines), dim), dtype='float32')
    id_to_tok = []
    tok_to_id = {}

    logger.info("Load glove embeddings as np array.")
    id = 0
    for line in tqdm.tqdm(glove_file_lines):
        split_line = line.split()
        word = split_line[0]
        emb_table[id] = np.array([float(val) for val in split_line[1:]])
        id_to_tok.append(word)
        tok_to_id[word] = id
        id += 1

    return {
        "emb_table": emb_table,
        "id2tok": id_to_tok,
        "tok2id": tok_to_id
    }


def get_glove_emb(download_only=False):
    """Download default pretrained glove embeddings and return a dict.

    We use the 300-dimensional model trained on Wikipedia 2014 + Gigaword 5.
    See https://nlp.stanford.edu/projects/glove/

    Args:
        download_only (bool): set True to only download. (Returns None)

    Returns:
        (dict): a dict of GloVe word embedding model.
            "emb_table": a numpy array of size(N, 300)
            "id2tok": a list of strings.
            "tok2id": a dict that maps word (string) to its id.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common")
    if not os.path.exists(os.path.join(data_dir, "glove.6B.300d.txt")):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls["default-glove-embeddings"])

    if download_only:
        return None
    return load_glove_model(os.path.join(data_dir, "glove.6B.300d.txt"), 300)


def get_stopwords():
    """Download default stopword words.

    Returns:
        ([str]): a list of strings.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common")
    download_file(subdir=os.path.join(data_dir),
                  **downloadable_resource_urls["default-stopwords"])

    with open(os.path.join(data_dir, "stopwords.txt")) as f:
        stopwords = f.readlines()
    stopwords = [x.strip().lower() for x in stopwords]
    return stopwords


def get_nltk_data():
    """Download nltk data to ``<fibber_root_dir>/nltk_data``."""
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common", "nltk_data", "tokenizers")
    if not os.path.exists(os.path.join(data_dir, "punkt")):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls["nltk-punkt"])

    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common", "nltk_data", "corpora")
    if not os.path.exists(os.path.join(data_dir, "stopwords")):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls["nltk_stopwords"])


def get_universal_sentence_encoder():
    """Download pretrained universal sentence encoder.

    Returns:
        (str): directory of the downloaded model.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common", "tfhub_pretrained",
                            "universal-sentence-encoder-large_5")
    if not os.path.exists(data_dir):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls["universal-sentence-encoder"])

    return data_dir


def get_corenlp():
    """Download stanford corenlp package.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common")
    if not os.path.exists(os.path.join(data_dir, "stanford-corenlp-4.1.0")):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls["stanford-corenlp"])


def get_transformers(name):
    """Download pretrained transformer models.

    Args:
        name (str): the name of the pretrained models. options are ``["bert-base-cased",
            "bert-base-uncased", "gpt2-medium"]``.

    Returns:
        (str): directory of the downloaded model.
    """
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "common", "transformers_pretrained")
    if not os.path.exists(os.path.join(data_dir, name)):
        download_file(subdir=os.path.join(data_dir),
                      **downloadable_resource_urls[name])

    return os.path.join(data_dir, name)


def get_bert_clf_demo():
    """Download the pretrained classifier for demo dataset."""
    data_dir = get_root_dir()
    data_dir = os.path.join(data_dir, "bert_clf")
    if not os.path.exists(os.path.join(data_dir, "demo")):
        download_file(subdir=data_dir,
                      **downloadable_resource_urls["bert-base-uncased-clf-demo"])


def get_bert_lm_demo(path="."):
    """Download the pretrained language model for demo dataset.

    Since this data is algorithm-specific, it is downloaded to ``path`` instead of
    ``<fibber_root_dir>``.
    """
    if not os.path.exists(os.path.join(path, "lm_all")):
        download_file(abs_path=path,
                      **downloadable_resource_urls["bert-base-uncased-lm-demo"])

    if not os.path.exists(os.path.join(path, "wordpiece_emb-demo-0500.pt")):
        download_file(abs_path=path,
                      **downloadable_resource_urls["wpe-demo"])



import json
import os

from fibber import log
from fibber.datasets.downloadable_datasets import downloadable_dataset_urls
from fibber.download_utils import download_file, get_root_dir

logger = log.setup_custom_logger(__name__)


def download_and_preprocess_mr():
    """preprocess raw movie review dataset to Fibber's JSON format."""
    root_dir = get_root_dir()
    dataset_dir = "datasets/mr/"

    download_file(subdir=os.path.join(dataset_dir, "raw"),
                  **downloadable_dataset_urls["mr-raw"])

    logger.info("Start processing data.")

    with open(os.path.join(root_dir, dataset_dir, "raw/rt-polaritydata/rt-polarity.neg"),
              encoding="utf-8", errors="ignore") as f:
        neg = f.readlines()

    with open(os.path.join(root_dir, dataset_dir, "raw/rt-polaritydata/rt-polarity.pos"),
              encoding="utf-8", errors="ignore") as f:
        pos = f.readlines()

    train = {
        "label_mapping": ["negative", "positive"],
        "cased": False,
        "paraphrase_field": "text0",
    }

    test = {
        "label_mapping": ["negative", "positive"],
        "cased": False,
        "paraphrase_field": "text0",
    }

    trainlist = []
    testlist = []

    for id, item in enumerate(neg):
        if id % 10 == 0:
            testlist.append({
                "label": 0,
                "text0": item.strip()})
        else:
            trainlist.append({
                "label": 0,
                "text0": item.strip()})

    for id, item in enumerate(pos):
        if id % 10 == 0:
            testlist.append({
                "label": 1,
                "text0": item.strip()})
        else:
            trainlist.append({
                "label": 1,
                "text0": item.strip()})

    train["data"] = trainlist
    test["data"] = testlist

    with open(os.path.join(root_dir, dataset_dir, "train.json"), "w") as f:
        json.dump(train, f, indent=2)

    with open(os.path.join(root_dir, dataset_dir, "test.json"), "w") as f:
        json.dump(test, f, indent=2)


if __name__ == "__main__":
    download_and_preprocess_mr()



"""This metric computes the embedding similarity using SBERT model."""

import numpy as np
import torch
from sentence_transformers import SentenceTransformer

from fibber import log
from fibber.metrics.metric_base import MetricBase

logger = log.setup_custom_logger(__name__)


class SBERTSemanticSimilarityMetric(MetricBase):
    """This metric computes the semantic similarity of two sentences using Sentence BERT model.

    By default the we use stsb-bert-base model.

    see `https://github.com/UKPLab/sentence-transformers` for more information.
    """

    def __init__(self, sbert_pretrained_model="stsb-bert-base", sbert_gpu_id=-1, **kargs):
        """Initialize sbert model."""
        super(SBERTSemanticSimilarityMetric, self).__init__()

        if sbert_gpu_id == -1:
            logger.warning("GPT2 metric is running on CPU.")
            self._device = torch.device("cpu")
        else:
            logger.info("GPT2 metric is running on GPU %d.", sbert_gpu_id)
            self._device = torch.device("cuda:%d" % sbert_gpu_id)

        logger.info("load sbert model.")

        self._model = SentenceTransformer(sbert_pretrained_model).to(self._device)

    def _get_emb(self, sentences):
        """Compute the embedding of sentences."""
        return self._model.encode(sentences)

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrase_list.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing the USE similarity metric for each paraphrase.
        """
        embs = self._get_emb([origin] + paraphrase_list)
        norm = np.linalg.norm(embs, axis=1)
        sim = np.sum(embs[0] * embs, axis=1) / norm[0] / norm
        return [float(x) for x in sim[1:]]

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the perplexity ratio.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record: ignored.
            paraphrase_field: ignored.
        """
        embs = self._get_emb([origin, paraphrase])
        return float(np.sum(embs[0] * embs[1]) / np.linalg.norm(embs[0]) / np.linalg.norm(embs[1]))



import numpy as np

from fibber import log
from fibber.datasets import builtin_datasets
from fibber.datasets.dataset_utils import get_dataset, verify_dataset
from fibber.metrics import MetricBundle
from fibber.paraphrase_strategies import (
    BertSamplingStrategy, IdentityStrategy, RandomStrategy, TextAttackStrategy)

logger = log.setup_custom_logger(__name__)


class Fibber(object):
    """Fibber is a unified interface for paraphrase strategies."""

    def __init__(self, arg_dict, dataset_name, strategy_name, trainset=None, testset=None,
                 output_dir=".", bert_clf_steps=5000):
        """Initialize

        Args:
            arg_dict (dict): a dict of hyper parameters for the MetricBundle and strategy.
            dataset_name (str): the name of the dataset.
            strategy_name (str): the strategy name.
            trainset (str): fibber dataset.
            testset (str): fibber testset.
            output_dir (str): directory to cache the strategy.
        """
        super(Fibber, self).__init__()

        # setup dataset
        if dataset_name in builtin_datasets:
            if trainset is not None or testset is not None:
                logger.error(("dataset name %d conflict with builtin dataset. "
                              "set trainset and testset to None.") % dataset_name)
                raise RuntimeError
            trainset, testset = get_dataset(dataset_name)
        else:
            verify_dataset(trainset)
            verify_dataset(testset)

        self._metric_bundle = MetricBundle(
            enable_bert_clf_prediction=True,
            use_gpu_id=arg_dict["use_gpu_id"],
            gpt2_gpu_id=arg_dict["gpt2_gpu_id"],
            bert_gpu_id=arg_dict["bert_gpu_id"],
            dataset_name=dataset_name,
            trainset=trainset, testset=testset,
            bert_clf_steps=bert_clf_steps)

        strategy_gpu_id = arg_dict["strategy_gpu_id"]
        if strategy_name == "RandomStrategy":
            self._strategy = RandomStrategy(
                arg_dict, dataset_name, strategy_gpu_id, output_dir, self._metric_bundle)
        if strategy_name == "IdentityStrategy":
            self._strategy = IdentityStrategy(
                arg_dict, dataset_name, strategy_gpu_id, output_dir, self._metric_bundle)
        if strategy_name == "TextAttackStrategy":
            self._strategy = TextAttackStrategy(
                arg_dict, dataset_name, strategy_gpu_id, output_dir, self._metric_bundle)
        if strategy_name == "BertSamplingStrategy":
            self._strategy = BertSamplingStrategy(
                arg_dict, dataset_name, strategy_gpu_id, output_dir, self._metric_bundle)
        if self._strategy is None:
            logger.error("unknown strategy name %s." % strategy_name)
            raise RuntimeError

        self._strategy.fit(trainset)

        self._trainset = trainset
        self._testset = testset

    def paraphrase(self, data_record, field_name="text0", n=20):
        """Paraphrase a given data record.

        Args:
            data_record (dict): data record to be paraphrased.
            field_name (str): select from ``["text0", "text1"]``
            n (int): number of paraphrases.

        Returns:
            * a list of str as paraphrased sentences.
            * a list of dict as corresponding metrics.
        """
        paraphrases = self._strategy.paraphrase_example(data_record, field_name, n)
        metrics = []
        for item in paraphrases:
            metrics.append(self._metric_bundle.measure_example(
                data_record[field_name], item, data_record, field_name))
        return data_record[field_name], paraphrases, metrics

    def paraphrase_a_random_sentence(self, n=20, from_testset=True):
        """Randomly pick one data, then paraphrase it.

        Args:
            n (int): number of paraphrases.
            from_testset (bool): if true, select data from test set, otherwise from training set.

        Returns:
            * a str as the original text.
            * a list of str as the paraphrased text.
            * a list of dict as corresponding metrics.
        """
        dataset = self._testset if from_testset else self._trainset
        field = dataset["paraphrase_field"]

        data_record = np.random.choice(dataset["data"])

        paraphrases, metrics = self.paraphrase(data_record, field_name=field, n=n)

        return data_record[field], paraphrases, metrics

    def get_metric_bundle(self):
        """"Get the metric bundle."""
        return self._metric_bundle



from fibber.datasets.dataset_utils import (
    DatasetForBert, get_dataset, get_demo_dataset, subsample_dataset, verify_dataset)

__all__ = [
    "get_dataset",
    "subsample_dataset",
    "verify_dataset",
    "DatasetForBert",
    "get_demo_dataset",
    "builtin_datasets"]

builtin_datasets = [
    "ag", "ag_no_title", "mr", "imdb", "yelp", "snli", "mnli", "mnli_mis"
]



from fibber.metrics.bert_classifier import BertClassifier
from fibber.metrics.edit_distance_metric import EditDistanceMetric
from fibber.metrics.glove_semantic_similarity_metric import GloVeSemanticSimilarityMetric
from fibber.metrics.gpt2_grammar_quality_metric import GPT2GrammarQualityMetric
from fibber.metrics.metric_base import MetricBase
from fibber.metrics.metric_utils import MetricBundle
from fibber.metrics.sbert_semantic_similarity_metric import SBERTSemanticSimilarityMetric
from fibber.metrics.use_semantic_similarity_metric import USESemanticSimilarityMetric

__all__ = [
    "BertClassifier",
    "EditDistanceMetric",
    "GloVeSemanticSimilarityMetric",
    "GPT2GrammarQualityMetric",
    "MetricBase",
    "USESemanticSimilarityMetric",
    "SBERTSemanticSimilarityMetric",
    "MetricBundle"]



from fibber.resources.resource_utils import (
    get_bert_clf_demo, get_bert_lm_demo, get_corenlp, get_glove_emb, get_nltk_data, get_stopwords,
    get_transformers, get_universal_sentence_encoder)


def download_all():
    get_nltk_data()
    get_glove_emb(download_only=True)
    get_transformers("bert-base-cased")
    get_transformers("bert-base-uncased")
    get_transformers("gpt2-medium")
    get_universal_sentence_encoder()
    get_corenlp()
    get_stopwords()


def download_resources_for_demo(demo_path="."):
    get_nltk_data()
    get_glove_emb(download_only=True)
    get_transformers("gpt2-medium")
    get_transformers("bert-base-uncased")
    get_universal_sentence_encoder()
    get_stopwords()

    get_bert_clf_demo()
    get_bert_lm_demo(demo_path)


if __name__ == "__main__":
    download_all()



"""This module defines customized metric aggregation functions."""

import itertools
import math
from multiprocessing import Pool

import numpy as np

from fibber.metrics.edit_distance_metric import EditDistanceMetric
from fibber.metrics.metric_utils import DIRECTION_HIGHER_BETTER, DIRECTION_LOWER_BETTER


def paraphrase_classification_accuracy_agg_fn_constructor(
        gpt2_ppl_threshold, use_sim_threshold, target_clf):
    """This function makes a aggregation function for the target classification metric.

    The aggregation function outputs the after attack accuracy of the BERT classifier.

    Args:
        gpt2_ppl_threshold (float): the threshold for ``GPT2GrammarQualityMetric`` metric. The
            adversarial example should have the perplexity ratio measured by GPT2 lower than
            this threshold.
        use_sim_threshold (float): the threshold for ``USESemanticSimilarityMetric`` metric. The
            adversarial example should have the USE similarity higher than this threshold.
        target_clf (str): the metric name of the target classifier.
    """

    def agg_fn(data_record):
        if data_record["original_text_metrics"][target_clf] != data_record["label"]:
            return 0
        for item in data_record["paraphrase_metrics"]:
            if (item[target_clf] != data_record["label"]
                    and item["GPT2GrammarQualityMetric"] < gpt2_ppl_threshold
                    and item["USESemanticSimilarityMetric"] > use_sim_threshold):
                return 0
        return 1

    return agg_fn


def editing_distance_element_worker(x):
    editing_distance_metric = EditDistanceMetric()
    return editing_distance_metric.measure_example(x[0], x[1])


pool = Pool(8)


def pairwise_editing_distance_fn(data_record):
    """Compute the average pairwise editing distance metric."""
    global pool
    paraphrases = None
    for k, v in data_record.items():
        if k in ["text0_paraphrases", "text1_paraphrases"]:
            paraphrases = v
            break

    assert paraphrases is not None

    tuples = list(itertools.combinations(paraphrases, 2))

    distance = pool.map(editing_distance_element_worker, tuples)
    return float(np.mean(distance))


def get_best_adv_by_sim(data_record, target_clf, gpt2_ppl_threshold=5, use_sim_threshold=0.85):
    """Find the adversarial example with best similarity.

    Args:
        data_record (dict): a data record with paraphrases and metrics.
        target_clf (str): the targeted classifier.
        gpt2_ppl_threshold (float): the gpt2 perplexity ration threshold for a legitimate
            adversarial example.
        use_sim_threshold (float): the USE cosine similarity threshold for a legitimate adversarial
            example.
    Returns:
         (dict): the metrics of the best adversarial example. None if no legitimate adversarial
            example is found.
    """
    if data_record["original_text_metrics"][target_clf] != data_record["label"]:
        return None
    best_score = -1
    best_metrics = None
    for metrics in data_record["paraphrase_metrics"]:
        if (metrics[target_clf] == data_record["label"]
                or metrics["GPT2GrammarQualityMetric"] > gpt2_ppl_threshold
                or metrics["USESemanticSimilarityMetric"] < use_sim_threshold):
            continue
        if metrics["USESemanticSimilarityMetric"] > best_score:
            best_score = metrics["USESemanticSimilarityMetric"]
            best_metrics = metrics
    return best_metrics


def get_best_adv_by_ppl(data_record, target_clf, gpt2_ppl_threshold=5, use_sim_threshold=0.85):
    """Find the adversarial example with lowest perplexity.

    Args:
        data_record (dict): a data record with paraphrases and metrics.
        target_clf (str): the targeted classifier.
        gpt2_ppl_threshold (float): the gpt2 perplexity ration threshold for a legitimate
            adversarial example.
        use_sim_threshold (float): the USE cosine similarity threshold for a legitimate adversarial
            example.
    Returns:
         (dict): the metrics of the best adversarial example. None if no legitimate adversarial
            example is found.
    """
    if data_record["original_text_metrics"][target_clf] != data_record["label"]:
        return None
    best_score = 1e8
    best_metrics = None
    for metrics in data_record["paraphrase_metrics"]:
        if (metrics[target_clf] == data_record["label"]
                or metrics["GPT2GrammarQualityMetric"] > gpt2_ppl_threshold
                or metrics["USESemanticSimilarityMetric"] < use_sim_threshold):
            continue
        if metrics["GPT2GrammarQualityMetric"] < best_score:
            best_score = metrics["GPT2GrammarQualityMetric"]
            best_metrics = metrics
    return best_metrics


def get_best_adv_metric_fn_constructor(get_best_adv_fn, metric_name, target_clf,
                                       gpt2_ppl_threshold=5, use_sim_threshold=0.85):
    """Returns an aggregation function that extracts the value of a specified metric for the best
    adversarial example.

    The aggregation function returns NaN if no legitimate adversarial example is found.

    Args:
        get_best_adv_fn (fn): a function that returns the metric dict of the best adversarial
            example.
        metric_name (str): a metric name.
        target_clf (str): the targeted classifier.
        gpt2_ppl_threshold (float): the gpt2 perplexity ration threshold for a legitimate
            adversarial example.
        use_sim_threshold (float): the USE cosine similarity threshold for a legitimate adversarial
            example.
    Returns:
        (fn): an aggregation function that takes data_record as an input.
    """
    def agg_fn(data_record):
        best_metrics = get_best_adv_fn(data_record, target_clf,
                                       gpt2_ppl_threshold, use_sim_threshold)
        if best_metrics is not None:
            return best_metrics[metric_name]
        return math.nan

    return agg_fn


def add_sentence_level_adversarial_attack_metrics(
        metric_bundle, gpt2_ppl_threshold, use_sim_threshold):
    """Add advanced aggregation functions related to adversarial attack to a specific metric
    bundle.

    Args:
        metric_bundle (MetricBundle): a metric bundle object to add aggregation functions.
        gpt2_ppl_threshold (float): the gpt2 perplexity ration threshold for a legitimate
            adversarial example.
        use_sim_threshold (float): the USE cosine similarity threshold for a legitimate adversarial
            example.
    """
    metric_bundle.add_advanced_aggregation_fn(
        "PairwiseEditDistance",
        pairwise_editing_distance_fn,
        DIRECTION_HIGHER_BETTER
    )

    for classifier_name in metric_bundle.get_classifier_names():
        metric_bundle.add_advanced_aggregation_fn(
            "3_%s_AfterAttackAccuracy_USE_%.2f_GPT_%.1f" % (
                classifier_name, use_sim_threshold, gpt2_ppl_threshold),
            paraphrase_classification_accuracy_agg_fn_constructor(
                gpt2_ppl_threshold=gpt2_ppl_threshold,
                use_sim_threshold=use_sim_threshold,
                target_clf=classifier_name),
            DIRECTION_LOWER_BETTER
        )

        for metric_name in metric_bundle.get_metric_names():
            metric_bundle.add_advanced_aggregation_fn(
                "%s_best_sim_adv_%s" % (classifier_name, metric_name),
                get_best_adv_metric_fn_constructor(
                    get_best_adv_by_sim, metric_name, metric_bundle.get_target_classifier_name()),
                metric_bundle.get_metric_direction(metric_name)
            )



import numpy as np

from fibber.paraphrase_strategies.strategy_base import StrategyBase


class RandomStrategy(StrategyBase):
    """Randomly shuffle words in a sentence to generate paraphrases."""

    __abbr__ = "rand"

    def paraphrase_example(self, data_record, field_name, n):
        text = data_record[field_name]
        tokens = text.split()
        ret = []
        for i in range(n):
            np.random.shuffle(tokens)
            ret.append(" ".join(tokens))

        return ret



"""Top-level package for fibber."""

__author__ = 'MIT Data To AI Lab'
__email__ = 'dailabmit@gmail.com'
__version__ = '0.2.2'

import os

import nltk


def get_root_dir():
    """Return ``~/.fibber``, the root dir for fibber to store datasets and common resources."""
    root_dir = os.path.join(os.path.expanduser('~'), ".fibber")
    os.makedirs(root_dir, exist_ok=True)
    return root_dir


# change cache directory
nltk.data.path += [os.path.join(get_root_dir(), "common", "nltk_data")]
os.environ['TRANSFORMERS_CACHE'] = os.path.join(
    get_root_dir(), "common", "transformers_pretrained")
os.environ['TFHUB_CACHE_DIR'] = os.path.join(get_root_dir(), "common", "tfhub_pretrained")
os.environ['CORENLP_HOME'] = os.path.join(get_root_dir(), "common", "stanford-corenlp-4.1.0")



from fibber.metrics.attack_aggregation_utils import (
    get_best_adv_by_ppl, get_best_adv_by_sim, pairwise_editing_distance_fn,
    paraphrase_classification_accuracy_agg_fn_constructor)


def make_data_record(label, origin_predict, paraphrase_ppl_list,
                     paraphrase_sim_list, paraphrase_pred_list, classifier):
    assert len(paraphrase_ppl_list) == len(paraphrase_sim_list)
    assert len(paraphrase_sim_list) == len(paraphrase_pred_list)

    return {
        "label": label,
        "original_text_metrics": {
            classifier: origin_predict
        },
        "paraphrase_metrics": [
            {
                "GPT2GrammarQualityMetric": ppl,
                "USESemanticSimilarityMetric": sim,
                classifier: pred
            } for ppl, sim, pred in zip(paraphrase_ppl_list,
                                        paraphrase_sim_list,
                                        paraphrase_pred_list)
        ]
    }


def test_paraphrase_classification_accuracy_agg_fn_constructor():
    classifier = "FooClassifier"
    agg_fn = paraphrase_classification_accuracy_agg_fn_constructor(2, 0.9, classifier)

    data_record = make_data_record(label=1, origin_predict=0,
                                   paraphrase_ppl_list=[],
                                   paraphrase_sim_list=[],
                                   paraphrase_pred_list=[],
                                   classifier=classifier)
    assert agg_fn(data_record) == 0

    data_record = make_data_record(label=1, origin_predict=1,
                                   paraphrase_ppl_list=[],
                                   paraphrase_sim_list=[],
                                   paraphrase_pred_list=[],
                                   classifier=classifier)
    assert agg_fn(data_record) == 1

    data_record = make_data_record(label=1, origin_predict=1,
                                   paraphrase_ppl_list=[5, 1.2, 1.1],
                                   paraphrase_sim_list=[0.98, 0.7, 0.95],
                                   paraphrase_pred_list=[2, 3, 1],
                                   classifier=classifier)
    assert agg_fn(data_record) == 1

    data_record = make_data_record(label=1, origin_predict=1,
                                   paraphrase_ppl_list=[5, 1.2, 1.1],
                                   paraphrase_sim_list=[0.98, 0.7, 0.95],
                                   paraphrase_pred_list=[2, 1, 2],
                                   classifier=classifier)
    assert agg_fn(data_record) == 0


def test_pairwise_editing_distance_fn():
    data_record = {
        "text0_paraphrases": [
            "aa bb cc",
            "aa cc ee",
            "aa bb"
        ]
    }
    assert abs(pairwise_editing_distance_fn(data_record) - 5 / 3) < 1e-6


def test_get_best_adv_by_sim():
    classifier = "FooClassifier"
    data_record = make_data_record(label=1, origin_predict=1,
                                   paraphrase_ppl_list=[5.1, 1.2, 1.1, 1.2, 1.3, 1.4],
                                   paraphrase_sim_list=[0.98, 0.7, 0.95, 0.92, 0.93, 0.94],
                                   paraphrase_pred_list=[2, 3, 1, 4, 5, 6],
                                   classifier=classifier)
    best_metric = get_best_adv_by_sim(data_record, classifier, 5, 0.85)
    assert best_metric["GPT2GrammarQualityMetric"] == 1.4
    assert best_metric["USESemanticSimilarityMetric"] == 0.94
    assert best_metric[classifier] == 6


def test_get_best_adv_by_ppl():
    classifier = "FooClassifier"
    data_record = make_data_record(label=1, origin_predict=1,
                                   paraphrase_ppl_list=[5.1, 1.2, 1.1, 1.2, 1.3, 1.4],
                                   paraphrase_sim_list=[0.98, 0.7, 0.95, 0.92, 0.93, 0.94],
                                   paraphrase_pred_list=[2, 3, 1, 4, 5, 6],
                                   classifier=classifier)
    best_metric = get_best_adv_by_ppl(data_record, classifier, 5, 0.85)
    assert best_metric["GPT2GrammarQualityMetric"] == 1.2
    assert best_metric["USESemanticSimilarityMetric"] == 0.92
    assert best_metric[classifier] == 4



import numpy as np

from fibber.metrics.metric_base import MetricBase


class ClassifierBase(MetricBase):
    """Base class for classifiers.

    All classifiers must be derived from this class.

    To implement a new classifier, you should at least overwrite the ``predict_dist_example``
    method. This method returns a predicted logits over classes.

    Some classifiers output label instead of distribution. In this case, you should return a
    one-hot vector.

    Some classifier may run more efficiently on a batch of data. In this case, you should overwrite
    the ``predict_dist_batch`` function. If you don't overwrite predict_dist_batch, it will compute
    the metric of paraphrase_list one by one.
    """

    def predict_dist_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Predict the log-probability distribution over classes for one example.

        Args:
            origin (str): the original text.
            paraphrase (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.array): a numpy array of size ``(num_labels)``.
        """
        raise NotImplementedError

    def predict_dist_batch(self, origin, paraphrase_list, data_record=None,
                           paraphrase_field="text0"):
        """Predict the log-probability distribution over classes for one batch.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.array): a numpy array of size ``(batch_size * num_labels)``.
        """
        ret = []
        for paraphrase in paraphrase_list:
            ret.append(
                self.predict_dist_example(origin, paraphrase, data_record, paraphrase_field))
        return np.asarray(ret)

    def predict_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Predict class label for one example.

        Args:
            origin (str): the original text.
            paraphrase (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.int): predicted label
        """
        return np.argmax(
            self.predict_dist_example(origin, paraphrase, data_record, paraphrase_field))

    def predict_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Predict class label for one example.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.array): predicted label as an numpy array of size ``(batch_size)``.
        """
        return np.argmax(
            self.predict_dist_batch(origin, paraphrase_list, data_record, paraphrase_field),
            axis=1)

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Predict class label for one example.

        Wrapper for ``predict_example``. Return type is changed from numpy.int to int.

        Args:
            origin (str): the original text.
            paraphrase (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (int): predicted label
        """
        return int(self.predict_example(origin, paraphrase, data_record, paraphrase_field))

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Predict class label for one batch.

         Wrapper for ``predict_batch``. Return type is changed from numpy.array to list of int.

         Args:
             origin (str): the original text.
             paraphrase_list (list): a set of paraphrase_list.
             data_record (dict): the corresponding data record of original text.
             paraphrase_field (str): the field name to paraphrase.

         Returns:
             ([int]): predicted label
         """
        return [int(x) for x in
                self.predict_batch(origin, paraphrase_list, data_record, paraphrase_field)]



import random
import re
import socket

from stanza.server import CoreNLPClient

from fibber import log, resources

logger = log.setup_custom_logger(__name__)


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(('localhost', port)) == 0


def generate_phrases_by_dfs(node, level, phrase_level, result):
    """DFS a parsing tree and generate all phrases in the tree."""
    if len(node.child) == 0:
        # if current node is leaf
        word = node.value.strip()
        if level <= phrase_level:
            result.append(word)
        return word

    phrase = ""
    for child in node.child:
        phrase += " " + generate_phrases_by_dfs(child, level + 1, phrase_level, result)
    phrase = " ".join(phrase.split())

    if level == phrase_level:
        result.append(phrase)
    return phrase


def preprocess(x):
    """Add a space after punctuation."""
    return re.sub(r"([.,?:!])([^\s])", r"\1 \2", x)


class TextParser(object):
    """TextParser uses stanford core nlp tool to parse text.

    Args:
        port(int): the port to launch a stanford core nlp client
    """

    def __init__(self, port=9000):
        resources.get_corenlp()
        while is_port_in_use(port):
            port += 1
        self._core_nlp_client = CoreNLPClient(
            annotators=['parse'], timeout=600000, memory='16G', be_quiet=True,
            endpoint="http://localhost:%d" % port)

    def _get_parse_tree(self, sentence):
        """Generate a parsing tree from a sentence."""
        annotation = self._core_nlp_client.annotate(sentence)
        if len(annotation.sentence) != 1:
            logger.warning("_get_parse_tree should take one sentence. but %s is given." % sentence)
        return annotation.sentence[0].parseTree

    def split_paragraph_to_sentences(self, paragraph):
        """Split a paragraph into a list of sentences.

        Args:
            paragraph (str): a paragraph of English text.
        Returns:
            ([str]): a list of sentences.
        """
        paragraph = preprocess(paragraph)
        annotation = self._core_nlp_client.annotate(paragraph)
        return [" ".join([token.word for token in sentence.token])
                for sentence in annotation.sentence]

    def get_phrases(self, sentence, phrase_level=2):
        """Split a sentence into phrases.

        Args:
            sentence (str): a sentence in English.
            phrase_level (int): larger value gives shorter phrases.
        Returns:
            ([str]): a list of phrases.
        """
        root = self._get_parse_tree(sentence)

        phrases = []
        generate_phrases_by_dfs(root, 0, phrase_level, phrases)
        return phrases

    def phrase_level_shuffle(self, paragraph, n):
        """For each sentence, randomly shuffle phrases in that sentence.

        Args:
            paraphrase (str): a paragraph in English.
            n (int): number of randomly shuffled paragraph to generate.
        Returns:
            ([str]): a list of ``n`` shuffled paragraphs.
        """
        paragraph = preprocess(paragraph)
        sentences = self.split_paragraph_to_sentences(paragraph)
        bins = []
        for sentence in sentences:
            bins.append(self.get_phrases(sentence))

        ret = []
        for i in range(n):
            sent = ""
            for bin in bins:
                bin_tmp = bin[:]
                random.shuffle(bin_tmp)
                sent += " " + " ".join(bin_tmp)

            ret.append(sent.strip())
        return ret



import json
import re

import pandas as pd
import tqdm

from fibber import log
from fibber.datasets.preprocess_utils import download_raw_and_preprocess

logger = log.setup_custom_logger(__name__)


REPLACE_TOKS = [
    ("#39;", "'"),
    ("#36;", "$"),
    ("&gt;", ">"),
    ("&lt;", "<"),
    ("\\$", "$"),
    ("quot;", "\""),
    ("\\", " "),
    ("#145;", "\""),
    ("#146;", "\""),
    ("#151;", "-")
]


def preprocess_ag_data(input_filename, output_filename,
                       include_title=True, include_author_media=True):
    """preprocess raw AG's news csv to Fibber's JSON format."""
    logger.info("Start preprocessing data, and save at %s.", output_filename)
    df = pd.read_csv(input_filename, header=None)

    data = {
        "label_mapping": ["World", "Sports", "Business", "Sci/Tech"],
        "cased": True,
        "paraphrase_field": "text0",
    }

    datalist = []
    for item in tqdm.tqdm(df.iterrows(), total=len(df)):
        y, title, text = item[1]
        y -= 1
        for u, v in REPLACE_TOKS:
            title = title.replace(u, v)
            text = text.replace(u, v)
            if not include_author_media:
                m = re.search(r"^[^\(\)]{0,50}?(\([^\(\)]{0,20}?\))? (-|--) ", text)
                if m is not None and len(m.group(0)) < len(text):
                    text = text[len(m.group(0)):]

        datalist.append({
            "label": y,
            "text0": (title + " . " + text) if include_title else text
        })

    data["data"] = datalist
    with open(output_filename, "w") as f:
        json.dump(data, f, indent=2)


def download_and_preprocess_ag():
    """Download and preprocess AG's news dataset. """
    download_raw_and_preprocess(
        dataset_name="ag",
        download_list=["ag-raw-train", "ag-raw-test"],
        preprocess_fn=preprocess_ag_data,
        preprocess_input_output_list=[
            ("raw/train.csv", "train.json"),
            ("raw/test.csv", "test.json")])

    download_raw_and_preprocess(
        dataset_name="ag_no_title",
        download_list=["ag-raw-train", "ag-raw-test"],
        preprocess_fn=lambda inp, out:
            preprocess_ag_data(inp, out, include_title=False, include_author_media=False),
        preprocess_input_output_list=[
            ("raw/train.csv", "train.json"),
            ("raw/test.csv", "test.json")])


if __name__ == "__main__":
    download_and_preprocess_ag()



"""This metric computes the cosine similarity between two sentences. The sentence embedding is
the sum of GloVe word embeddings."""


import numpy as np
from nltk import word_tokenize

from fibber import log
from fibber.metrics.metric_base import MetricBase
from fibber.resources import get_glove_emb, get_nltk_data, get_stopwords

logger = log.setup_custom_logger('glove_semantic_similairty')


def compute_emb(emb_table, tok_to_id, x):
    """Compute the sum of word embeddings for a sentence.

    Args:
        emb_table (np.array): the glove embedding table.
        tok_to_id (dict): a dict mapping strs to ints.
        x (str): text.

    Returns:
        (np.array): the sum of word embedding.
    """
    toks = word_tokenize(x)
    embs = []
    for item in toks:
        if item.lower() in tok_to_id:
            embs.append(emb_table[tok_to_id[item.lower()]])
    return np.sum(embs, axis=0)


def compute_emb_sim(emb_table, tok_to_id, x, y):
    """Compute the cosine similarity between two sentences. The sentence embedding is the sum
    of word embeddings.

    Args:
        emb_table (np.array): the glove embedding table.
        tok_to_id (dict): a dict mapping strs to ints.
        x (str): text.
        y (str): text.

    Returns:
        (float): the cosine similarity.
    """
    ex = compute_emb(emb_table, tok_to_id, x)
    ey = compute_emb(emb_table, tok_to_id, y)
    return ((ex * ey).sum()
            / (np.linalg.norm(ex) + 1e-8)
            / (np.linalg.norm(ey) + 1e-8))


class GloVeSemanticSimilarityMetric(MetricBase):
    """This metric computes the cosine similarity between two sentences."""

    def __init__(self, **kargs):
        """Initialize, load Glove embeddings."""
        super(GloVeSemanticSimilarityMetric, self).__init__()

        get_nltk_data()
        self._glove = get_glove_emb()
        stopwords = get_stopwords()
        logger.info("Glove embeddings and stopwords loaded.")

        for word in stopwords:
            word = word.lower().strip()
            if word in self._glove["tok2id"]:
                self._glove["emb_table"][self._glove["tok2id"][word], :] = 0

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the Glove cosine similarity between two sentences.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record: ignored.
            paraphrase_field: ignored.
        """
        return float(compute_emb_sim(self._glove["emb_table"], self._glove["tok2id"],
                                     origin, paraphrase))



"""This metric computes the perplexity ratio ppl(paraphrase) / ppl(original text).

The perplexity is estimated using GPT2 model. This metric can reveal the meaningfulness of a
sentence.
"""

import numpy as np
import torch
import torch.nn.functional as F
from transformers import GPT2LMHeadModel, GPT2TokenizerFast

from fibber import log, resources
from fibber.metrics.metric_base import MetricBase

logger = log.setup_custom_logger(__name__)


def make_input_output_pair(tokenizer, x):
    """Tokenize the text, then construct input and output for GPT2."""
    toks = tokenizer.encode(x, add_special_tokens=True)
    toks = toks[:200]
    return [tokenizer.bos_token_id] + toks[:-1], toks


def make_batch(toks_list):
    """Convert multiple text to a batch tensor."""
    n = len(toks_list)
    max_len = max([len(x) for x in toks_list])

    ids = np.zeros((n, max_len), dtype='int')
    mask = np.zeros((n, max_len), dtype='int')

    for i, item in enumerate(toks_list):
        ids[i, :len(item)] = np.asarray(item)
        mask[i, :len(item)] = 1

    return ids, mask


class GPT2GrammarQualityMetric(MetricBase):
    """This metric computes the perplexity of paraphrased text divided by the perplexity of
    original text. The perplexity is measured using GPT2 model.
    """

    def __init__(self, gpt2_pretrained_model="gpt2-medium", gpt2_gpu_id=-1, **kargs):
        """Initialize GPT2 model."""
        super(GPT2GrammarQualityMetric, self).__init__()

        logger.info("load gpt2 model.")
        self._tokenizer = GPT2TokenizerFast.from_pretrained(
            resources.get_transformers(gpt2_pretrained_model))
        if gpt2_gpu_id == -1:
            logger.warning("GPT2 metric is running on CPU.")
            self._device = torch.device("cpu")
        else:
            logger.info("GPT2 metric is running on GPU %d.", gpt2_gpu_id)
            self._device = torch.device("cuda:%d" % gpt2_gpu_id)

        self._model = GPT2LMHeadModel.from_pretrained(
            resources.get_transformers(gpt2_pretrained_model)).to(
            self._device)

    def _get_ppl(self, sentences):
        """Compute the perplexity of sentences."""
        input_output = [make_input_output_pair(self._tokenizer, x) for x in sentences]

        input, output = zip(*input_output)

        toks_input, mask = make_batch(input)
        toks_output, _ = make_batch(output)

        mask = torch.tensor(mask).to(self._device)
        toks_input = torch.tensor(toks_input).to(self._device)
        toks_output = torch.tensor(toks_output).to(self._device)

        with torch.no_grad():
            logits = self._model(toks_input, attention_mask=mask)[0]

        logpw = torch.gather(F.log_softmax(logits, dim=-1), dim=-1,
                             index=toks_output.unsqueeze(dim=2)).squeeze(dim=2)
        ppl = torch.exp(-(logpw * mask).sum(dim=1) / mask.sum(dim=1))
        ppl = ppl.detach().cpu().numpy()
        return ppl

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrase_list.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing the USE similarity metric for each paraphrase.
        """
        ppls = self._get_ppl([origin] + paraphrase_list)
        res = ppls[1:] / ppls[0]
        return [float(x) for x in res]

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the perplexity ratio.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record: ignored.
            paraphrase_field: ignored.
        """
        ppl = self._get_ppl([origin, paraphrase])
        return float(ppl[1] / ppl[0])



import pandas as pd

from fibber.benchmark.benchmark_utils import load_detailed_result, update_overview_result
from fibber.metrics.metric_utils import DIRECTION_HIGHER_BETTER, DIRECTION_LOWER_BETTER

DATASET_NAME_COL = "0_dataset_name"
STRATEGY_NAME_COL = "1_paraphrase_strategy_name"


def make_overview():
    """Generate overview table from detailed table."""
    detailed_df = load_detailed_result()

    # verify detailed result
    for group_info, item in detailed_df.groupby([DATASET_NAME_COL, STRATEGY_NAME_COL]):
        assert len(item) == 1, (
            "Detailed results contains multiple runs for %s on %s." % (
                group_info[1], group_info[0]))

    col_for_num_win = []

    for col_name in detailed_df.columns.tolist():
        if col_name.endswith(DIRECTION_HIGHER_BETTER):
            col_for_num_win.append((col_name, DIRECTION_HIGHER_BETTER))
        if col_name.endswith(DIRECTION_LOWER_BETTER):
            col_for_num_win.append((col_name, DIRECTION_LOWER_BETTER))

    results = {}
    for rid, item in detailed_df.iterrows():
        if item[STRATEGY_NAME_COL] not in results:
            model_name = item[STRATEGY_NAME_COL]
            tmp = dict()

            tmp[STRATEGY_NAME_COL] = item[STRATEGY_NAME_COL]
            for col_name, _ in col_for_num_win:
                tmp[col_name] = 0

            results[model_name] = tmp

    for group_name, group in detailed_df.groupby(DATASET_NAME_COL):
        for _, r1 in group.iterrows():
            for _, r2 in group.iterrows():
                for column_name, direction in col_for_num_win:
                    if ((direction == DIRECTION_HIGHER_BETTER
                         and r1[column_name] > r2[column_name])
                            or (direction == DIRECTION_LOWER_BETTER
                                and r1[column_name] < r2[column_name])):
                        results[r1[STRATEGY_NAME_COL]][column_name] += 1

    update_overview_result(pd.DataFrame(list(results.values())))


if __name__ == "__main__":
    make_overview()



import json

import pandas as pd
import tqdm

from fibber import log
from fibber.datasets.preprocess_utils import download_raw_and_preprocess

logger = log.setup_custom_logger(__name__)

REPLACE_TOKS = [
    ("\\\"", "\""),
    ("\\n", " ")
]


def preprocess_yelp_data(input_filename, output_filename):
    """preprocess Yelp dataset to Fibber's JSON format."""
    logger.info("Start preprocessing data, and save at %s.", output_filename)

    df = pd.read_csv(input_filename, header=None)

    data = {
        "label_mapping": ["negative", "positive"],
        "cased": True,
        "paraphrase_field": "text0",
    }

    datalist = []
    for item in tqdm.tqdm(df.iterrows(), total=len(df)):
        y, text = item[1]
        y -= 1
        for u, v in REPLACE_TOKS:
            text = text.replace(u, v)

        datalist.append({
            "label": y,
            "text0": text
        })

    data["data"] = datalist
    with open(output_filename, "w") as f:
        json.dump(data, f, indent=2)


def download_and_preprocess_yelp():
    """Download and preprocess Yelp dataset."""
    download_raw_and_preprocess(
        dataset_name="yelp",
        download_list=["yelp-raw"],
        preprocess_fn=preprocess_yelp_data,
        preprocess_input_output_list=[
            ("raw/yelp_review_polarity_csv/train.csv", "train.json"),
            ("raw/yelp_review_polarity_csv/test.csv", "test.json")])


if __name__ == "__main__":
    download_and_preprocess_yelp()



import datetime
import json

import numpy as np
import pandas as pd
import tqdm

from fibber import log
from fibber.metrics.bert_classifier import BertClassifier
from fibber.metrics.classifier_base import ClassifierBase
from fibber.metrics.edit_distance_metric import EditDistanceMetric
from fibber.metrics.glove_semantic_similarity_metric import GloVeSemanticSimilarityMetric
from fibber.metrics.gpt2_grammar_quality_metric import GPT2GrammarQualityMetric
from fibber.metrics.metric_base import MetricBase
from fibber.metrics.use_semantic_similarity_metric import USESemanticSimilarityMetric

logger = log.setup_custom_logger(__name__)


DIRECTION_HIGHER_BETTER = "(↑)"
DIRECTION_LOWER_BETTER = "(↓)"
DIRECTION_UNKNOWN = "(x)"


class MetricBundle(object):
    """MetricBundle can help easily initialize and compute multiple metrics."""

    def __init__(self,
                 enable_edit_distance=True,
                 enable_use_semantic_similarity=True,
                 enable_glove_semantic_similarity=True,
                 enable_gpt2_grammar_quality=True,
                 enable_bert_clf_prediction=False,
                 **kargs):
        """Initialize various metrics.

        Args:
            enable_edit_distance (bool): whether to use editing distance in the bundle.
            enable_use_semantic_similarity (bool): whether to use Universal sentence encoder to
                compute sentence similarity
            enable_glove_semantic_similarity (bool): whether to use Glove embeddings to compute
                sentence similarity.
            enable_gpt2_grammar_quality (bool): whether to use GPT2 to compute sentence quality.
            enable_bert_clf_prediction (bool): whether to include BERT classifier prediction in
                metrics.
            kargs: arguments for metrics. kargs will be passed to all metrics.
        """
        super(MetricBundle, self).__init__()

        self._metrics = {}
        self._classifiers = {}
        self._target_clf = None

        self._advanced_aggregation_fn = {}

        if enable_edit_distance:
            self.add_metric(EditDistanceMetric(**kargs), DIRECTION_HIGHER_BETTER)
        if enable_use_semantic_similarity:
            self.add_metric(USESemanticSimilarityMetric(**kargs), DIRECTION_HIGHER_BETTER)
        if enable_glove_semantic_similarity:
            self.add_metric(GloVeSemanticSimilarityMetric(**kargs), DIRECTION_HIGHER_BETTER)
        if enable_gpt2_grammar_quality:
            self.add_metric(GPT2GrammarQualityMetric(**kargs), DIRECTION_LOWER_BETTER)
        if enable_bert_clf_prediction:
            self.add_classifier(BertClassifier(**kargs), set_target_clf=True)

    def add_metric(self, metric, direction):
        """Add a customized metric to metric bundle.

        Args:
            metric (MetricBase): the metric object to add.
            direction (str): choose from ``DIRECTION_HIGHER_BETTER``, ``DIRECTION_HIGHER_BETTER``
                and ``DIRECTION_UNKNOWN``.
        """
        if not isinstance(metric, MetricBase):
            logger.error("%s is not an instance of MetricBase.", str(metric))
            raise RuntimeError
        if str(metric) in self._metrics:
            logger.error("Duplicate metric %s.", str(metric))
            raise RuntimeError
        self._metrics[str(metric)] = (metric, direction)

    def get_metric(self, metric_name):
        """Returns a metric in the bundle using the metric name.

        Metric name is the class name of a metric.

        Raises assertion error if metric is not found.

        Args:
            metric_name: the name of the matric.
        Returns:
            (object): a metric object.
        """
        assert metric_name in self._metrics
        return self._metrics[metric_name][0]

    def get_metric_direction(self, metric_name):
        """Returns the direction of a metric.

        Metric name is the class name of a metric.

        Raises assertion error if metric is not found.

        Args:
            metric_name: the name of the matric.
        Returns:
            (object): a metric object.
        """
        assert metric_name in self._metrics
        return self._metrics[metric_name][1]

    def get_metric_names(self):
        """"Returns all metric names in this metric bundle.

        Returns:
            list of str
        """
        return list(self._metrics.keys())

    def add_classifier(self, classifier_metric, set_target_clf=False):
        """Set a target classifier to attack.

        Args:
            classifier_metric (ClassifierBase): A classifier metric to be added.
            set_target_clf (bool): whether to set this classifier metric as target classifier.
        """
        assert isinstance(classifier_metric, ClassifierBase)
        self._classifiers[str(classifier_metric)] = classifier_metric
        self._target_clf = str(classifier_metric)
        if set_target_clf:
            self.set_target_classifier_by_name(str(classifier_metric))

    def get_classifier(self, classifier_name):
        """Returns the classifier in current metric bundle.

        Args:
            classifier_name (str): the name of the requested classifier.
        """
        return self._classifiers[classifier_name]

    def get_classifier_names(self):
        return list(self._classifiers.keys())

    def set_target_classifier_by_name(self, classifier_name):
        """Set a target classifier to attack.

        Args:
            classifier_name (str): set a classifier as target classifier.
        """
        assert isinstance(classifier_name, str)
        assert classifier_name in self._classifiers
        self._target_clf = classifier_name

    def get_target_classifier_name(self):
        """Return the name of the target classifier."""
        return self._target_clf

    def get_target_classifier(self):
        """Returns the classifier for attack."""
        return self.get_classifier(self.get_target_classifier_name())

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the results of all metrics in the bundle for one pair of text.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record (str): the data record.
            paraphrase_field (str): choose from "text0", "text1".

        Returns:
            (dict): a dict with metric name as key.
        """
        ret = {}
        for name in self.get_metric_names():
            metric = self.get_metric(name)
            ret[name] = metric.measure_example(origin, paraphrase, data_record, paraphrase_field)
        for name in self.get_classifier_names():
            classifier = self.get_classifier(name)
            ret[name] = classifier.measure_example(
                origin, paraphrase, data_record, paraphrase_field)
        return ret

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrase_list.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing dict of metrics for each paraphrase.
        """
        ret = [{} for i in range(len(paraphrase_list))]
        for name in self.get_metric_names():
            metric = self.get_metric(name)
            result = metric.measure_batch(origin, paraphrase_list, data_record, paraphrase_field)
            for i in range(len(paraphrase_list)):
                ret[i][name] = result[i]
        for name in self.get_classifier_names():
            classifier = self.get_classifier(name)
            result = classifier.measure_batch(origin, paraphrase_list,
                                              data_record, paraphrase_field)
            for i in range(len(paraphrase_list)):
                ret[i][name] = result[i]
        return ret

    def measure_dataset(self, results, output_filename):
        """Compute the all metrics for results on a dataset.

        Args:
            results (dict): A fibber dataset with paraphrase_list.
            output_filename (str): A json filename to store results and metrics.

        Returns:
            (dict): the results dict with ``original_text_metrics`` and ``paraphrase_metrics``
                added.
        """
        last_output_save_time = -1
        logger.info("Start measuring.")
        paraphrase_field = results["paraphrase_field"]

        for data_record in tqdm.tqdm(results["data"]):
            data_record_tmp = dict([(k, v) for k, v in data_record.items()
                                    if "_paraphrases" not in k])

            # Run metrics on original text
            data_record["original_text_metrics"] = self.measure_example(
                data_record[paraphrase_field], data_record[paraphrase_field],
                data_record_tmp, paraphrase_field)

            # Run metrics on paraphrased text
            data_record["paraphrase_metrics"] = self.measure_batch(
                data_record[paraphrase_field], data_record[paraphrase_field + "_paraphrases"],
                data_record_tmp, paraphrase_field)

            # save tmp output every 30 seconds
            if datetime.datetime.now().timestamp() - last_output_save_time > 30:
                with open(output_filename, "w") as f:
                    json.dump(results, f, indent=2)
                datetime.datetime.now().timestamp()

        with open(output_filename, "w") as f:
            json.dump(results, f, indent=2)

        return results

    def add_advanced_aggregation_fn(self, aggregation_name, aggregation_fn, direction):
        """Add advanced aggregation function.

        Some aggregation function can aggregate multiple metrics, these aggregation functions
        can be added here.

        Args:
            aggregation_name (str): the name of the aggregation.
            aggregation_fn (fn): an aggregation function that takes ``data_record`` as arg.
            direction (str): chose from DIRECTION_HIGHER_BETTER, DIRECTION_LOWER_BETTER,
                and DIRECTION_UNKNOWN.
        """
        if aggregation_name in self._advanced_aggregation_fn:
            logger.error("Duplicate advanced aggregation function %s.", aggregation_name)
            raise RuntimeError
        self._advanced_aggregation_fn[aggregation_name] = (aggregation_fn, direction)

    def get_advanced_aggregation_fn_names(self):
        return list(self._advanced_aggregation_fn.keys())

    def get_advanced_aggregation_fn_direction(self, aggregation_name):
        return self._advanced_aggregation_fn[aggregation_name][1]

    def get_advanced_aggregation_fn(self, aggregation_name):
        return self._advanced_aggregation_fn[aggregation_name][0]

    def aggregate_metrics(self, dataset_name, paraphrase_strategy_name, experiment_name, results):
        """Aggregate paraphrase metrics on a dataset to a summary and store in a dict.

        Args:
            dataset_name (str): the name of the dataset.
            paraphrase_strategy_name (str): the name of the paraphrase strategy.
            experiment_name (str): the name of the experiment.
            results (dict): the fibber dataset with paraphrases and metrics. The return value of
                ``compute_metrics``.

        Returns:
            (dict): the aggregated metrics.
        """
        aggregated_result = pd.DataFrame()
        for data_record in results["data"]:
            aggregate_result_tmp = {}
            metric_df = pd.DataFrame(data_record["paraphrase_metrics"])
            for metric_name in metric_df.columns.tolist():
                if metric_name in self.get_classifier_names():
                    continue
                metric_value = metric_df[metric_name].values
                direction = self.get_metric_direction(metric_name)

                # aggregate mean
                agg_value = np.mean(metric_value)
                aggregate_result_tmp[metric_name + direction] = float(agg_value)

                # aggregate std
                agg_value = np.std(metric_value)
                aggregate_result_tmp[metric_name + "_std"] = float(agg_value)

            aggregate_result_tmp["ParaphrasesPerExample"] = len(data_record["paraphrase_metrics"])

            for agg_name in self.get_advanced_aggregation_fn_names():
                agg_fn = self.get_advanced_aggregation_fn(agg_name)
                direction = self.get_advanced_aggregation_fn_direction(agg_name)
                aggregate_result_tmp[agg_name + direction] = agg_fn(data_record)

            aggregated_result = aggregated_result.append(aggregate_result_tmp, ignore_index=True)

        aggregated_result = dict(aggregated_result.mean(skipna=True))
        # hack column order by adding 0
        aggregated_result["0_dataset_name"] = dataset_name
        aggregated_result["1_paraphrase_strategy_name"] = paraphrase_strategy_name
        aggregated_result["2_experiment_name"] = experiment_name

        return aggregated_result



import os

import numpy as np
import torch
import torch.nn.functional as F
import tqdm
from torch.utils.tensorboard import SummaryWriter
from transformers import BertForMaskedLM

from fibber import log, resources
from fibber.datasets import DatasetForBert
from fibber.metrics.bert_classifier import get_optimizer

logger = log.setup_custom_logger(__name__)


def write_summary(stats, summary, global_step):
    """Save langauge model training summary."""
    summary.add_scalar("attack/loss_lm", np.mean(stats["lm_loss"]), global_step)
    summary.add_scalar("attack/error_lm", 1 - stats["lm_correct"] / stats["lm_total"], global_step)


def new_stats():
    """Create a new stats dict."""
    return {
        "lm_total": 0,
        "lm_correct": 0,
        "lm_loss": [],
    }


def compute_lm_loss(lm_model, seq, mask, tok_type, lm_label, stats):
    """Compute masked language model training loss.

    Args:
        lm_model (transformers.BertForMaskedLM): a BERT language model.
        seq (torch.Tensor): an int tensor of size (batch_size, length) representing the word
            pieces.
        mask (torch.Tensor): an int tensor of size (batch_size, length) representing the attention
            mask.
        tok_type (torch.Tensor): an int tensor of size (batch_size, length) representing the
            token type id.
        lm_label (torch.Tensor): an int tensor of size (batch_size, length) representing the label
            for each position. Use -100 if the loss is not computed for that position.
        stats (dist): a dictionary storing training stats.

    Returns:
        (torch.Scalar) a scalar loss value.
    """
    lm_hid = lm_model.bert(seq, mask, tok_type)[0]
    lm_hid = torch.masked_select(lm_hid, lm_label.gt(0).unsqueeze(2)).view(-1, lm_hid.size(2))
    logits = lm_model.cls(lm_hid)

    with torch.no_grad():
        lm_label_squeeze = torch.masked_select(lm_label, lm_label.gt(0))

    lm_loss = F.cross_entropy(logits, lm_label_squeeze)

    # lm_label is -100 for unmasked token or padding toks.
    stats["lm_total"] += (lm_label_squeeze > 0).int().sum().detach().cpu().numpy()
    stats["lm_correct"] += (logits.argmax(dim=1).eq(lm_label_squeeze)
                            .float().sum().detach().cpu().numpy())
    stats["lm_loss"].append(lm_loss.detach().cpu().numpy())

    return lm_loss


def get_lm(output_dir, trainset, filter, device, lm_steps=5000, lm_bs=32,
           lm_opt="adamw", lm_lr=0.0001, lm_decay=0.01,
           lm_period_summary=100, lm_period_save=5000):
    """Returns a finetuned BERT language model on a given dataset.

    The language model will be stored at ``<output_dir>/lm_all`` if filter is -1, or
    ``<output_dir>/lm_filter_?`` if filter is not -1.

    If filter is not -1. The pretrained langauge model will first be pretrained on the while
    dataset, then it will be finetuned on the data excluding the filter category.

    Args:
        output_dir (str): a directory to store pretrained language model.
        trainset (DatasetForBert): the training set for finetune the language model.
        filter (int): a category to exclude from finetuning.
        device (torch.Device): a device to train the model.
        lm_steps (int): finetuning steps.
        lm_bs (int): finetuning batch size.
        lm_opt (str): optimzer name. choose from ["sgd", "adam", "adamW"].
        lm_lr (float): learning rate.
        lm_decay (float): weight decay for the optimizer.
        lm_period_summary (int): number of steps to write training summary.
        lm_period_save (int): number of steps to save the finetuned model.
    Returns:
        (BertForMaskedLM): a finetuned language model.
    """
    if filter == -1:
        output_dir_t = os.path.join(output_dir, "lm_all")
    else:
        output_dir_t = os.path.join(output_dir, "lm_filter_%d" % filter)

    summary = SummaryWriter(output_dir_t + "/summary")

    if trainset["cased"]:
        model_init = "bert-base-cased"
    else:
        model_init = "bert-base-uncased"

    ckpt_path_pattern = output_dir_t + "/checkpoint-%04dk"
    ckpt_path = ckpt_path_pattern % (lm_steps // 1000)

    if os.path.exists(ckpt_path):
        logger.info("Language model <%s> exists.", ckpt_path)
        return BertForMaskedLM.from_pretrained(ckpt_path).eval()

    if filter == -1:
        lm_model = BertForMaskedLM.from_pretrained(resources.get_transformers(model_init))
        lm_model.train()
    else:
        lm_model = get_lm(output_dir, trainset, -1, device, lm_steps)
        lm_model.train()
    lm_model.to(device)

    dataset = DatasetForBert(trainset, model_init, lm_bs, exclude=filter, masked_lm=True)
    dataloader = torch.utils.data.DataLoader(
        dataset, batch_size=None, num_workers=2)

    params = list(lm_model.parameters())
    opt, sche = get_optimizer(lm_opt, lm_lr, lm_decay, lm_steps, params)

    global_step = 0
    stats = new_stats()
    for seq, mask, tok_type, label, lm_label in tqdm.tqdm(
            dataloader, total=lm_steps):
        opt.zero_grad()

        global_step += 1

        seq = seq.to(device)
        mask = mask.to(device)
        tok_type = tok_type.to(device)
        label = label.to(device)
        lm_label = lm_label.to(device)

        lm_loss = compute_lm_loss(
            lm_model, seq, mask, tok_type, lm_label, stats)
        lm_loss.backward()

        opt.step()
        sche.step()

        if global_step % lm_period_summary == 0:
            write_summary(stats, summary, global_step)
            stats = new_stats()

        if global_step % lm_period_save == 0 or global_step == lm_steps:
            lm_model.to(torch.device("cpu")).eval()
            lm_model.save_pretrained(ckpt_path_pattern % (global_step // 1000))
            lm_model.to(device)

        if global_step >= lm_steps:
            break

    lm_model.eval()
    lm_model.to(torch.device("cpu"))
    return lm_model



class MetricBase(object):
    """Base class for Metrics.

    All metrics should be derived from this class.

    To implement a new metric, you should at least overwrite the ``measure_example`` method.

    The simplest metric can be directly computed from a pair of text, in this case, the metric
    can use the ``origin`` and ``paraphrase`` args directly.

    Other metrics need more information from the data record. For example, ``text0``, ``text1``,
    or ``label``. Thus the ``data_record`` and ``paraphrase_field`` are also provided as args.

    Some metrics may run more efficiently on a batch of data. In this case, you should overwrite
    the ``measure_batch`` function. If you don't overwrite batch_call, it will compute the metric
    of paraphrase_list one by one.
    """

    def __init__(self, **kargs):
        super(MetricBase, self).__init__()

    def __repr__(self):
        return self.__class__.__name__

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrase_list.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing the metric for each paraphrase.
        """
        ret = []
        for paraphrase in paraphrase_list:
            ret.append(self.measure_example(origin, paraphrase, data_record, paraphrase_field))
        return ret

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        raise NotImplementedError()



from fibber.paraphrase_strategies.bert_sampling_strategy import BertSamplingStrategy
from fibber.paraphrase_strategies.identity_strategy import IdentityStrategy
from fibber.paraphrase_strategies.random_strategy import RandomStrategy
from fibber.paraphrase_strategies.strategy_base import StrategyBase
from fibber.paraphrase_strategies.textattack_strategy import TextAttackStrategy

__all__ = ["IdentityStrategy", "RandomStrategy", "StrategyBase", "BertSamplingStrategy",
           "TextAttackStrategy"]



import logging

G_FORMATTER = logging.Formatter(
    fmt="%(asctime)s - %(levelname)s - %(module)s - %(message)s")


def setup_custom_logger(name, level="INFO"):
    """Get a logger."""
    logger = logging.getLogger(name)

    handler = logging.StreamHandler()
    handler.setFormatter(G_FORMATTER)
    logger.addHandler(handler)

    logger.setLevel(getattr(logging, level))
    return logger


def add_file_handler(logger, filename):
    """Add file handler to a logger."""
    file_handler = logging.FileHandler(filename)
    file_handler.setFormatter(G_FORMATTER)
    logger.root.addHandler(file_handler)


def remove_logger_tf_handler(logger):
    """Remove all handlers except file handler.

    This function can clean up the mess caused by tensorflow_hub.
    """
    logger.root.handlers = [item for item in logger.root.handlers
                            if isinstance(item, logging.FileHandler)]



import math
import re

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F
from transformers import BertForMaskedLM, BertTokenizerFast

from fibber import log, resources
from fibber.paraphrase_strategies.bert_sampling_utils_lm import get_lm
from fibber.paraphrase_strategies.bert_sampling_utils_text_parser import TextParser
from fibber.paraphrase_strategies.bert_sampling_utils_wpe import get_wordpiece_emb
from fibber.paraphrase_strategies.strategy_base import StrategyBase

logger = log.setup_custom_logger(__name__)

POST_PROCESSING_PATTERN = [
    (r"\s+n\s", "n "),
    (r"\s*'\s*t\s", "'t "),
    (r"\s*'\s*s\s", "'s "),
    (r"\s*'\s*ve\s", "'ve "),
    (r"\s*'\s*ll\s", "'ll "),
    (r"\s*n't\s", "n't "),
    (r"- -", "--"),
    (r"\s*([\.,?!])", r"\1"),
    (r"\s+-\s+", "-"),
]

PRE_PROCESSING_PATTERN = [
    (r"can't\s", r" cannot "),
    (r"won't\s", r" will not "),
    (r"n't\s", r" not "),
    (r"'ll\s", r" will "),
    (r"'ve\s", r" have "),
]

AUTO_SENTENCE_LEN_THRESHOLD = 50  # 50 words


def process_text(text, patterns):
    """Processing the text using regex patterns.

    Args:
        text (str): the str to be post processed.
    """
    for pattern in patterns:
        text = re.sub(pattern[0], pattern[1], text)
    return text


def tostring(tokenizer, seq):
    """Convert a sequence of word ids to a sentence. The post prossing is applied.

    Args:
        tokenizer (transformers.BertTokenizer): a BERT tokenizer.
        seq (list): a list-like sequence of word ids.
    """
    return process_text(tokenizer.decode(seq), POST_PROCESSING_PATTERN)


def sample_word_from_logits(logits, temperature=1., top_k=0):
    """Sample a word from a distribution.

    Args:
        logits (torch.Tensor): tensor of logits with size ``(batch_size, vocab_size)``.
        temperature (float): the temperature of softmax. The PMF is
            ``softmax(logits/temperature)``.
        top_k (int): if ``k>0``, only sample from the top k most probable words.
    """
    logits = logits / temperature

    if top_k > 0:
        kth_vals, kth_idx = logits.topk(top_k, dim=-1)
        dist = torch.distributions.categorical.Categorical(logits=kth_vals)
        idx = kth_idx.gather(dim=1, index=dist.sample().unsqueeze(-1)).squeeze(-1)
    else:
        dist = torch.distributions.categorical.Categorical(logits=logits)
        idx = dist.sample().squeeze(-1)
    return idx


def all_accept_criteria(candidate_ids, stats, **kargs):
    """Always accept proposed words.

    Args:
        candidate_ids (torch.Tensor): proposed word ids in this sampling step with
            size ``(batch_size, pos_ed-pos_st)``.
        stats (dict): a dict to keep track the accept rate.

    Returns:
        (np.array, None)
            np.array is the same as candidate_ids.
            None means this criteria does not have any state.
    """
    stats["accept"] += len(candidate_ids)
    stats["all"] += len(candidate_ids)
    return candidate_ids, None


def use_criteria_score(origin, paraphrases, use_metric, use_threshold, use_weight):
    """Estimate the score of a sentence using USE.

    Args:
        origin (str): original sentence.
        paraphrases ([str]): a list of paraphrase_list.
        use_metric (USESemanticSimilarityMetric): a universal sentence encoder metric object.
        use_threshold (float): the universal sentence encoder similarity threshold.
        use_weight (float): the weight parameter for the criteria.

    Returns:
        (np.array): a numpy array of size ``(batch_size,)``. All entries ``<=0``.
    """
    if use_weight == 0:
        return np.zeros(len(paraphrases), dtype="float32")
    use_semantic_similarity = use_metric.measure_batch(origin, paraphrases)
    return -use_weight * (
        np.maximum(use_threshold - np.asarray(use_semantic_similarity), 0) ** 2)


def gpt2_criteria_score(origin, paraphrases, gpt2_metric, gpt2_weight):
    """Estimate the score of a sentence using USE.

    Args:
        origin (str): original sentence.
        paraphrases ([str]): a list of paraphrase_list.
        gpt2_metric (GPT2GrammarQualityMetric): a GPT2GrammarQualityMetric metric object.
        gpt2_weight (float): the weight parameter for the criteria.

    Returns:
        (np.array): a numpy array of size ``(batch_size,)``. All entries ``<=0``.
    """
    if gpt2_weight == 0:
        return np.zeros(len(paraphrases), dtype="float32")
    gpt2_ppl_ratio = gpt2_metric.measure_batch(origin, paraphrases)
    return -gpt2_weight * (np.maximum(gpt2_ppl_ratio, 0) ** 2)


def bert_criteria_score(origin, paraphrases, data_record, field_name, bert_metric, bert_weight):
    if bert_weight == 0:
        return np.zeros(len(paraphrases), dtype="float32")

    dist = bert_metric.predict_dist_batch(origin, paraphrases, data_record, field_name)
    label = data_record["label"]
    correct_prob = (dist[:, label]).copy()
    dist[:, label] = -1e8
    incorrect_prob = np.max(dist, axis=1)
    return -bert_weight * np.maximum(correct_prob - incorrect_prob, 0)


def joint_weighted_criteria(
        tokenizer, data_record, field_name, origin, batch_tensor,
        pos_st, pos_ed, previous_ids, candidate_ids, use_metric, use_threshold, use_weight,
        bert_metric, bert_weight, gpt2_metric, gpt2_weight, burnin_weight, stats, state,
        device, **kargs):
    """Accept or reject candidate word using the joint weighted criteria.

    Args:
        tokenizer (transformers.BertTokenizer): a bert tokenizer.
        data_record (dict): the data record dict.
        field_name (str): the field to rewritten.
        origin (str): original text. Same as ``data_record[field_name]``.
        batch_tensor (torch.Tensor): tensor of a batch of text with size ``(batch_size, L)``.
        pos_st (int): the start position of sampling (include).
        pos_ed (int): the end position of sampling (exclude).
        previous_ids (torch.Tensor): word ids before current step of sampling with
            size ``(batch_size, pos_ed-pos_st)``.
        candidate_ids (torch.Tensor): proposed word ids in this sampling step with
            size ``(batch_size, pos_ed-pos_st)``.
        use_metric (USESemanticSimilarityMetric): a universal sentence encoder metric object.
        use_threshold (float): the universal sentence encoder similarity threshold.
        use_weight (float): the weight for USE criteria score.
        bert_metric (BertClassifier): a BertClassifier metric.
        bert_weight (float): the weight for BERT criteria score.
        gpt2_metric (GPT2GrammarQualityMetric): a GPT2GrammarQualityMetric metric.
        gpt2_weight (float): the weight for GPT2 criteria score.
        burnin_weight (float): the discount factor.
        stats (dict): a dict to keep track the accept rate.
        state (np.array): the state is criteria score from the previous iteration.
        device (torch.Device): the device that batch_tensor is on.
    Returns:
        (np.array, np.array)
            a 2-D int array of size ``batch_size, pos_ed - pos_st``. Each row ``i`` is
                either ``previous_ids[i, :]`` if rejected, or ``candidate_ids[i, :]`` if accepted.
            a 1-D float array of criteria score.
    """
    if burnin_weight == 0:
        return candidate_ids

    def compute_criteria_score(fill_ids):
        batch_tensor[:, pos_st:pos_ed] = fill_ids
        paraphrases = [tostring(tokenizer, x) for x in batch_tensor.detach().cpu().numpy()]
        return (gpt2_criteria_score(origin=origin, paraphrases=paraphrases,
                                    gpt2_metric=gpt2_metric, gpt2_weight=gpt2_weight)
                + use_criteria_score(origin=origin, paraphrases=paraphrases, use_metric=use_metric,
                                     use_weight=use_weight, use_threshold=use_threshold)
                + bert_criteria_score(origin=origin, paraphrases=paraphrases,
                                      data_record=data_record, field_name=field_name,
                                      bert_metric=bert_metric,
                                      bert_weight=bert_weight)) * burnin_weight

    if state is not None:
        previous_criteria_score = state
    else:
        previous_criteria_score = compute_criteria_score(previous_ids)

    candidate_criteria_score = compute_criteria_score(candidate_ids)

    alpha = np.exp(candidate_criteria_score - previous_criteria_score)

    accept = np.asarray(np.random.rand(len(alpha)) < alpha, dtype="int32")

    stats["accept"] += np.sum(accept)
    stats["all"] += len(accept)
    state = candidate_criteria_score * accept + previous_criteria_score * (1 - accept)
    accept = torch.tensor(accept).to(device)
    ids = candidate_ids * accept.reshape(-1, 1) + previous_ids * (1 - accept.reshape(-1, 1))
    return ids, state


def none_constraint(**kargs):
    return 0


def allow_list_constraint(allow_list, **kargs):
    return -1e6 * (1 - allow_list)


def wpe_constraint(target_emb, word_embs, batch_tensor, pos,
                   wpe_threshold, wpe_weight, **kargs):
    current_emb = (word_embs(batch_tensor[:, 1:-1]).sum(dim=1)
                   - word_embs(batch_tensor[:, pos]))
    candidate_emb = current_emb[:, None, :] + word_embs.weight.data[None, :, :]
    dis = F.cosine_similarity(candidate_emb, target_emb[:, None, :], dim=2)
    dis = (wpe_threshold - dis).clamp_(min=0)
    return -wpe_weight * dis * dis


class BertSamplingStrategy(StrategyBase):
    __abbr__ = "bs"
    __hyperparameters__ = [
        ("batch_size", int, 50, "the batch size in sampling."),
        ("window_size", int, 3, "the block sampling window size."),
        ("burnin_steps", int, 100, "number of burnin steps."),
        ("sampling_steps", int, 200, "number of sampling steps (including) burnin."),
        ("top_k", int, 100, "sample from top k words after burnin. Use 0 for all words."),
        ("temperature", float, 1., "the softmax temperature for sampling."),
        ("use_threshold", float, 0.95, "the threshold for USE similarity."),
        ("use_weight", float, 1000, "the smoothing parameter for USE similarity."),
        ("wpe_threshold", float, 1.00, "the threshold for USE similarity."),
        ("wpe_weight", float, 10000, "the smoothing parameter for USE similarity."),
        ("burnin_enforcing_schedule", str, "1",
            ("the schedule decides how much additional "
             "constraint is added. options are [linear, 0, 1].")),
        ("accept_criteria", str, "joint_weighted_criteria", (
            "select an accept criteria for candidate words from "
            "[all, joint_weighted_criteria].")),
        ("enforcing_dist", str, "wpe", ("select an additional constraint for candidate "
                                        "words from [none, allow_list, wpe].")),
        ("burnin_criteria_schedule", str, "1", ("the schedule decides how strict the criteria is "
                                                "used. options are [linear, 0, 1].")),
        ("seed_option", str, "origin", ("the option for seed sentences in generation. "
                                        "choose from [origin, auto].")),
        ("split_sentence", str, "0", "split paragraph to sentence. options are [0, 1, auto]."),
        ("stanza_port", int, 9000, "stanza port"),
        ("lm_option", str, "finetune", "choose from [pretrain, finetune, adv]."),
        ("lm_steps", int, 5000, "lm training steps."),
        ("clf_weight", float, 1, "weight for the clf score in the criteria."),
        ("gpt2_weight", float, 10, "the smoothing parameter for gpt2."),
    ]

    def __repr__(self):
        return "%s-constraint_%s-criteria_%s" % (
            self.__class__.__name__,
            self._strategy_config["enforcing_dist"],
            self._strategy_config["accept_criteria"])

    def fit(self, trainset):
        # load BERT language model.
        logger.info("Load bert language model for BertSamplingStrategy.")

        if trainset["cased"]:
            model_init = "bert-base-cased"
        else:
            model_init = "bert-base-uncased"
        self._tokenizer = BertTokenizerFast.from_pretrained(
            resources.get_transformers(model_init), do_lower_case="uncased" in model_init)
        self._tokenizer.do_lower_case = True if "uncased" in model_init else False

        if self._strategy_config["lm_option"] == "pretrain":
            self._bert_lm = BertForMaskedLM.from_pretrained(
                resources.get_transformers(model_init)).to(self._device)
            self._bert_lm.eval()
            for item in self._bert_lm.parameters():
                item.requires_grad = False
        elif self._strategy_config["lm_option"] == "finetune":
            self._bert_lm = get_lm(
                self._output_dir, trainset, -1, self._device, self._strategy_config["lm_steps"])
            self._bert_lm.eval()
            self._bert_lm.to(self._device)
            for item in self._bert_lm.parameters():
                item.requires_grad = False
        elif self._strategy_config["lm_option"] == "adv":
            self._bert_lms = []
            for i in range(len(trainset["label_mapping"])):
                lm = get_lm(
                    self._output_dir, trainset, i, self._device, self._strategy_config["lm_steps"])
                lm.eval()
                for item in lm.parameters():
                    item.requires_grad = False
                self._bert_lms.append(lm)
        else:
            assert 0

        # Load useful metrics
        self._use_metric = self._metric_bundle.get_metric("USESemanticSimilarityMetric")
        self._clf_metric = self._metric_bundle.get_target_classifier()
        self._gpt2_metric = self._metric_bundle.get_metric("GPT2GrammarQualityMetric")

        # load word piece embeddings.
        wpe = get_wordpiece_emb(self._output_dir, self._dataset_name, trainset, self._device)
        self._word_embs = nn.Embedding(self._tokenizer.vocab_size, 300)
        self._word_embs.weight.data = torch.tensor(wpe.T).float()
        self._word_embs = self._word_embs.to(self._device)
        self._word_embs.eval()
        for item in self._word_embs.parameters():
            item.requires_grad = False

        # config _decision_fn and _enforcing_dist_fn
        if self._strategy_config["accept_criteria"] == "all":
            self._decision_fn = all_accept_criteria
        elif self._strategy_config["accept_criteria"] == "joint_weighted_criteria":
            self._decision_fn = joint_weighted_criteria
        else:
            assert 0

        if self._strategy_config["enforcing_dist"] == "none":
            self._enforcing_dist_fn = none_constraint
        elif self._strategy_config["enforcing_dist"] == "wpe":
            self._enforcing_dist_fn = wpe_constraint
        elif self._strategy_config["enforcing_dist"] == "allow_list":
            self._enforcing_dist_fn = allow_list_constraint
        else:
            assert 0

        # load text parser
        if (self._strategy_config["seed_option"] != "origin"
                or self._strategy_config["split_sentence"] != "0"):
            self._text_parser = TextParser(self._strategy_config["stanza_port"])
        else:
            self._text_parser = None

        self._stats = {
            "all": 0,
            "accept": 0
        }

    def _parallel_sequential_generation(self, seed, batch_size, burnin_steps, sampling_steps,
                                        field_name, data_record):
        if self._strategy_config["seed_option"] == "origin":
            seq = ["[CLS]"] + self._tokenizer.tokenize(seed) + ["[SEP]"]
            batch_tensor = torch.tensor(
                [self._tokenizer.convert_tokens_to_ids(seq)] * batch_size).to(self._device)
            tensor_len = len(seq)
        elif self._strategy_config["seed_option"] == "auto":
            seeds = self._text_parser.phrase_level_shuffle(seed, batch_size)
            seq = [self._tokenizer.tokenize(x) for x in seeds]
            tensor_len = max([len(x) for x in seq])
            seq = [["[CLS]"] + x + ["[MASK]"] * (tensor_len - len(x)) + ["[SEP]"] for x in seq]
            tensor_len += 2
            batch_tensor = torch.tensor(
                [self._tokenizer.convert_tokens_to_ids(x) for x in seq]).to(self._device)
        else:
            assert 0

        if field_name == "text1":
            context_seq = ["[CLS]"] + self._tokenizer.tokenize(data_record["text0"])
            context_tensor = torch.tensor(
                [self._tokenizer.convert_tokens_to_ids(context_seq)] * batch_size
            ).to(self._device)
            context_len = len(context_seq)
            batch_tensor[:, 0] = self._tokenizer.sep_token_id
        else:
            context_tensor = None

        target_emb = self._word_embs(batch_tensor[:, 1:-1]).sum(dim=1)

        allow_list = F.one_hot(batch_tensor[0][1:-1], self._tokenizer.vocab_size).sum(dim=0)

        decision_fn_state = None

        for ii in range(sampling_steps):
            pos_st = np.random.randint(1, tensor_len - 1)
            pos_ed = min(pos_st + self._strategy_config["window_size"], tensor_len - 1)

            previous_ids = batch_tensor[:, pos_st:pos_ed].clone()
            batch_tensor[:, pos_st:pos_ed] = self._tokenizer.mask_token_id

            sample_order = np.arange(pos_st, pos_ed)
            np.random.shuffle(sample_order)
            for pos in sample_order:
                if field_name == "text1":
                    batch_tensor_tmp = torch.cat([context_tensor, batch_tensor], dim=1)
                    tok_type_tensor_tmp = torch.cat([
                        torch.zeros_like(context_tensor),
                        torch.ones_like(batch_tensor)], dim=1)
                    logits_lm = self._bert_lm(
                        batch_tensor_tmp,
                        token_type_ids=tok_type_tensor_tmp)[0][:, context_len + pos]
                    logits_lm[:, self._tokenizer.sep_token_id] = -1e8
                else:
                    logits_lm = self._bert_lm(batch_tensor)[0][:, pos]

                logits_enforcing = self._enforcing_dist_fn(
                    # for wpe constraint
                    target_emb=target_emb,
                    word_embs=self._word_embs,
                    batch_tensor=batch_tensor,
                    pos=pos,
                    wpe_threshold=self._strategy_config["wpe_threshold"],
                    wpe_weight=self._strategy_config["wpe_weight"],
                    # for naive constraint
                    allow_list=allow_list
                )

                if ii < burnin_steps:
                    if self._strategy_config["burnin_enforcing_schedule"] == "0":
                        logits_joint = logits_lm
                    elif self._strategy_config["burnin_enforcing_schedule"] == "1":
                        logits_joint = logits_lm + logits_enforcing
                    elif self._strategy_config["burnin_enforcing_schedule"] == "linear":
                        logits_joint = logits_lm + (
                            ii / burnin_steps) * logits_enforcing
                    else:
                        assert 0
                else:
                    logits_joint = logits_lm + logits_enforcing

                top_k = self._strategy_config["top_k"] if (
                    ii >= burnin_steps) else 0

                candidate_ids = sample_word_from_logits(
                    logits_joint, top_k=top_k, temperature=self._strategy_config["temperature"])
                batch_tensor[:, pos] = candidate_ids

            candidate_ids = batch_tensor[:, pos_st:pos_ed].clone()

            if ii < burnin_steps:
                if self._strategy_config["burnin_criteria_schedule"] == "0":
                    decision_fn_burnin_weight = 0
                elif self._strategy_config["burnin_criteria_schedule"] == "1":
                    decision_fn_burnin_weight = 1
                elif self._strategy_config["burnin_criteria_schedule"] == "linear":
                    decision_fn_burnin_weight = ii / burnin_steps
                else:
                    assert 0
            else:
                pass

            final_ids, decision_fn_state = self._decision_fn(
                tokenizer=self._tokenizer, data_record=data_record, field_name=field_name,
                origin=data_record[field_name], batch_tensor=batch_tensor,
                pos_st=pos_st, pos_ed=pos_ed, previous_ids=previous_ids,
                candidate_ids=candidate_ids, use_metric=self._use_metric,
                use_threshold=self._strategy_config["use_threshold"],
                use_weight=self._strategy_config["use_weight"],
                bert_metric=self._clf_metric, bert_weight=self._strategy_config["clf_weight"],
                gpt2_metric=self._gpt2_metric, gpt2_weight=self._strategy_config["gpt2_weight"],
                burnin_weight=decision_fn_burnin_weight, stats=self._stats,
                state=decision_fn_state, device=self._device)

            batch_tensor[:, pos_st:pos_ed] = final_ids

        return [tostring(self._tokenizer, x[1:-1]) for x in batch_tensor.detach().cpu().numpy()]

    def paraphrase_example(self, data_record, field_name, n):
        if self._strategy_config["lm_option"] == "adv":
            self._bert_lm = self._bert_lms[data_record["label"]]
            self._bert_lm.to(self._device)

        clipped_text = " ".join(data_record[field_name].split()[:200])
        clipped_text = process_text(clipped_text, PRE_PROCESSING_PATTERN)
        batch_size = self._strategy_config["batch_size"]

        sentences = []
        n_batches = math.ceil(n / batch_size)
        last_batch_size = n % batch_size
        if last_batch_size == 0:
            last_batch_size = batch_size

        for id in range(n_batches):
            burnin_steps = self._strategy_config["burnin_steps"]
            sampling_steps = self._strategy_config["sampling_steps"]

            if self._strategy_config["split_sentence"] == "0":
                batch = self._parallel_sequential_generation(
                    clipped_text,
                    batch_size if id != n_batches - 1 else last_batch_size,
                    burnin_steps,
                    sampling_steps,
                    field_name, data_record)
                sentences += batch
            elif self._strategy_config["split_sentence"] in ["1", "auto"]:
                splitted_text_ori = self._text_parser.split_paragraph_to_sentences(
                    clipped_text)

                if self._strategy_config["split_sentence"] == "auto":
                    splitted_text = []
                    current_text = ""
                    for s in splitted_text_ori:
                        current_text += " " + s
                        if len(current_text.split()) > AUTO_SENTENCE_LEN_THRESHOLD:
                            splitted_text.append(current_text)
                            current_text = ""
                    if len(current_text.split()) > 0:
                        splitted_text.append(current_text)

                    burnin_steps = self._strategy_config["burnin_steps"] // len(splitted_text)
                    sampling_steps = self._strategy_config["sampling_steps"] // len(splitted_text)
                elif self._strategy_config["split_sentence"] == "1":
                    splitted_text = splitted_text_ori
                    burnin_steps = self._strategy_config["burnin_steps"] // len(splitted_text)
                    sampling_steps = self._strategy_config["sampling_steps"] // len(splitted_text)
                else:
                    assert 0

                batch_res = [""] * (batch_size if id != n_batches - 1 else last_batch_size)
                for text in splitted_text:
                    batch = self._parallel_sequential_generation(
                        text, batch_size if id != n_batches - 1 else last_batch_size,
                        burnin_steps,
                        sampling_steps,
                        field_name, data_record)

                    batch_res = [(x + " " + y).strip() for (x, y) in zip(batch_res, batch)]
                sentences += batch_res
            else:
                assert 0

        assert len(sentences) == n

        if self._strategy_config["lm_option"] == "adv":
            self._bert_lm.to(torch.device("cpu"))

        logger.info("Aggregated accept rate: %.2lf%%.",
                    self._stats["accept"] / self._stats["all"] * 100)
        return sentences[:n]



"""This metric computes the cosine similarity between two sentences. The sentence embedding is
the universal sentence encoder."""
import os

import numpy as np
import tensorflow as tf
import tensorflow_hub as hub

from fibber import log, resources
from fibber.metrics.metric_base import MetricBase

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
tf.get_logger().setLevel("ERROR")
logger = log.setup_custom_logger(__name__)
log.remove_logger_tf_handler(logger)    # tensorflow_hub mess up the python logging


def config_tf_gpu(gpu_id):
    """Configure tensorflow to use a specific GPU.

    Args:
        gpu_id (int): the gpu id. Set -1 to use CPU.
    """
    if tf.__version__ >= "2.3.0":
        gpus = tf.config.list_physical_devices(device_type="GPU")
        gpus = [item for item in gpus if item.name.endswith("GPU:%d" % gpu_id)]
        tf.config.set_visible_devices(gpus, device_type="GPU")
        for device in gpus:
            tf.config.experimental.set_memory_growth(device, True)
    else:
        gpus = tf.config.experimental.list_physical_devices(device_type="GPU")
        gpus = [item for item in gpus if item.name.endswith("GPU:%d" % gpu_id)]
        tf.config.experimental.set_visible_devices(gpus, device_type="GPU")
        for device in gpus:
            tf.config.experimental.set_memory_growth(device, True)


class USESemanticSimilarityMetric(MetricBase):
    """This metric uses universal sentence encoder to measure the semantic similarity of
    two sentences."""

    def __init__(self, use_gpu_id=-1, **kargs):
        """Initialize universal sentence encoder."""
        super(USESemanticSimilarityMetric, self).__init__()
        logger.info("load universal sentence encoder")
        config_tf_gpu(use_gpu_id)
        if use_gpu_id == -1:
            logger.warning("Universal sentence encoder is using CPU.")
        else:
            logger.info("Universal sentence encoder metric is using GPU %d.", use_gpu_id)
        self.model = hub.load(resources.get_universal_sentence_encoder())
        log.remove_logger_tf_handler(logger)   # tensorflow_hub mess up the python logging

    def measure_batch(self, origin, paraphrase_list, data_record=None, paraphrase_field="text0"):
        """Measure the metric on a batch of paraphrase_list.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (list): a list containing the USE similarity metric for each paraphrase.
        """
        origin = " ".join(origin.split()[:200])
        paraphrase_list = [" ".join(x.split()[:200]) for x in paraphrase_list]
        embs = self.model([origin] + paraphrase_list).numpy()

        norm = np.linalg.norm(embs, axis=1)
        sim = np.sum(embs[0] * embs, axis=1) / norm[0] / norm
        assert abs(sim[0] - 1) < 1e-4
        return [float(x) for x in sim[1:]]

    def measure_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Compute the cosine similarity between the embedding of original text and paraphrased
        text.

        Args:
            origin (str): original text.
            paraphrase (str): paraphrased text.
            data_record: ignored.
            paraphrase_field: ignored.
        """
        origin = " ".join(origin.split()[:200])
        paraphrase = " ".join(paraphrase.split()[:200])
        embs = self.model([origin, paraphrase]).numpy()
        return float(np.sum(embs[0] * embs[1]) / np.linalg.norm(embs[0]) / np.linalg.norm(embs[1]))



"""This module contains utility functions to download files to the root dir ``~/.fibber``."""

import hashlib
import os
import tarfile
import zipfile

from tensorflow.keras.utils import get_file as tf_get_file

from fibber import get_root_dir, log

logger = log.setup_custom_logger(__name__)


def check_file_md5(filename, md5):
    """Check if the md5 of a given file is correct.

    Args:
        filename (str): a filename.
        md5 (str): expected md5 hash value.
    Returns:
        (bool): Return True if md5 matches
    """
    if not os.path.exists(filename):
        return False
    hash_md5 = hashlib.md5()
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest() == md5


def download_file(filename, url, md5, subdir=None, untar=False, unzip=False, abs_path=None):
    """Download file from a given url.

    This downloads a file to ``<fibber_root_dir>/subdir``. If the file already exists and the md5
    matches, using the existing file.

    Args:
        filename (str): filename as a string.
        url (str): the url to download the file.
        md5 (str): the md5 checksum of the file.
        subdir (str): the subdir to save the file. Dir will be created if not exists.
        untar (bool): whether to untar the file.
        unzip (bool): whether to unzip the file.
        abs_path (str): a folder to download files. (ignore fibber_root_dir)
    """
    target_dir = get_root_dir()
    if subdir is not None:
        target_dir = os.path.join(target_dir, subdir)
    if abs_path is not None:
        target_dir = abs_path
    os.makedirs(target_dir, exist_ok=True)
    target_file_absolute_path = os.path.join(target_dir, filename)

    if (os.path.exists(target_file_absolute_path)
            and check_file_md5(target_file_absolute_path, md5)):
        logger.info("Load %s from cache. md5 checksum is correct.", filename)
        if untar:
            my_tar = tarfile.open(target_file_absolute_path)
            my_tar.extractall(target_dir)
            my_tar.close()
        if unzip:
            my_zip = zipfile.ZipFile(target_file_absolute_path, "r")
            my_zip.extractall(target_dir)
            my_zip.close()
    else:
        logger.info("Download %s to %s", filename, target_dir)
        tf_get_file(filename, origin=url, cache_subdir="",
                    file_hash=md5, extract=untar or unzip, cache_dir=target_dir)



from fibber.paraphrase_strategies.strategy_base import StrategyBase


class IdentityStrategy(StrategyBase):
    """A baseline paraphrase strategy. Just return the original sentence."""

    __abbr__ = "id"

    def paraphrase_example(self, data_record, field_name, n):
        return [data_record[field_name]]



"""This metric outputs the BERT classifier prediction of the paraphrased text."""

import os

import numpy as np
import torch
import torch.nn.functional as F
import tqdm
import transformers
from torch.utils.tensorboard import SummaryWriter
from transformers import BertForSequenceClassification, BertTokenizerFast

from fibber import get_root_dir, log, resources
from fibber.datasets import DatasetForBert
from fibber.metrics.classifier_base import ClassifierBase

logger = log.setup_custom_logger(__name__)


def get_optimizer(optimizer_name, lr, decay, train_step, params, warmup=1000):
    """Create an optimizer and schedule of learning rate for parameters.

    Args:
        optimizer_name (str): choose from ``["adam", "sgd", "adamw"]``.
        lr (float): learning rate.
        decay (float): weight decay.
        train_step (int): number of training steps.
        params (list): a list of parameters in the model.
        warmup (int): number of warm up steps.

    Returns:
        A torch optimizer and a scheduler.
    """
    if optimizer_name == "adam":
        opt = torch.optim.Adam(params=params, lr=lr, weight_decay=decay)
    elif optimizer_name == "sgd":
        opt = torch.optim.SGD(params=params, lr=lr, weight_decay=decay)
    elif optimizer_name == "adamw":
        opt = torch.optim.AdamW(params=params, lr=lr, weight_decay=decay)
    else:
        assert 0, "unkown optimizer"

    sche = transformers.get_linear_schedule_with_warmup(
        opt, num_warmup_steps=warmup, num_training_steps=train_step)

    return opt, sche


def run_evaluate(model, dataloader_iter, eval_steps, summary, global_step, device):
    """Evaluate a model and add error rate and validation loss to Tensorboard.

    Args:
        model (transformers.BertForSequenceClassification): a BERT classification model.
        dataloader_iter (torch.IterableDataset): an iterator of a torch.IterableDataset.
        eval_steps (int): number of training steps.
        summary (torch.utils.tensorboard.SummaryWriter): a Tensorboard SummaryWriter object.
        global_step (int): current training steps.
        device (torch.Device): the device where the model in running on.
    """
    model.eval()

    correct, count = 0, 0
    loss_list = []

    for v_step in range(eval_steps):
        seq, mask, tok_type, label = next(dataloader_iter)
        seq = seq.to(device)
        mask = mask.to(device)
        tok_type = tok_type.to(device)
        label = label.to(device)

        with torch.no_grad():
            outputs = model(seq, mask, tok_type, labels=label)
            loss, logits = outputs[:2]
            loss_list.append(loss.detach().cpu().numpy())
            count += seq.size(0)
            correct += (logits.argmax(dim=1).eq(label)
                        .float().sum().detach().cpu().numpy())

    summary.add_scalar("clf_val/error_rate", 1 - correct / count, global_step)
    summary.add_scalar("clf_val/loss", np.mean(loss_list), global_step)

    model.train()


def load_or_train_bert_clf(model_init,
                           dataset_name,
                           trainset,
                           testset,
                           bert_clf_steps,
                           bert_clf_bs,
                           bert_clf_lr,
                           bert_clf_optimizer,
                           bert_clf_weight_decay,
                           bert_clf_period_summary,
                           bert_clf_period_val,
                           bert_clf_period_save,
                           bert_clf_val_steps,
                           device):
    """Train BERT classification model on a dataset.

    The trained model will be stored at ``<fibber_root_dir>/bert_clf/<dataset_name>/``. If there's
    a saved model, load and return the model. Otherwise, train the model using the given data.

    Args:
        model_init (str): pretrained model name. Choose from ``["bert-base-cased",
            "bert-base-uncased", "bert-large-cased", "bert-large-uncased"]``.
        dataset_name (str): the name of the dataset. This is also the dir to save trained model.
        trainset (dict): a fibber dataset.
        testset (dict): a fibber dataset.
        bert_clf_steps (int): steps to train a classifier.
        bert_clf_bs (int): the batch size.
        bert_clf_lr (float): the learning rate.
        bert_clf_optimizer (str): the optimizer name.
        bert_clf_weight_decay (float): the weight decay.
        bert_clf_period_summary (int): the period in steps to write training summary.
        bert_clf_period_val (int): the period in steps to run validation and write validation
            summary.
        bert_clf_period_save (int): the period in steps to save current model.
        bert_clf_val_steps (int): number of batched in each validation.
        device (torch.Device): the device to run the model.

    Returns:
        (transformers.BertForSequenceClassification): a torch BERT model.
    """
    model_dir = os.path.join(get_root_dir(), "bert_clf", dataset_name)
    ckpt_path = os.path.join(model_dir, model_init + "-%04dk" %
                             (bert_clf_steps // 1000))

    if os.path.exists(ckpt_path):
        logger.info("Load BERT classifier from %s.", ckpt_path)
        model = BertForSequenceClassification.from_pretrained(ckpt_path)
        model.eval()
        model.to(device)
        return model

    num_labels = len(trainset["label_mapping"])

    model = BertForSequenceClassification.from_pretrained(
        resources.get_transformers(model_init), num_labels=num_labels).to(device)
    model.train()

    logger.info("Use %s tokenizer and classifier.", model_init)
    logger.info("Num labels: %s", num_labels)

    summary = SummaryWriter(os.path.join(model_dir, "summary"))

    dataloader = torch.utils.data.DataLoader(
        DatasetForBert(trainset, model_init, bert_clf_bs), batch_size=None, num_workers=2)

    dataloader_val = torch.utils.data.DataLoader(
        DatasetForBert(testset, model_init, bert_clf_bs), batch_size=None, num_workers=1)
    dataloader_val_iter = iter(dataloader_val)

    params = model.parameters()

    opt, sche = get_optimizer(
        bert_clf_optimizer, bert_clf_lr, bert_clf_weight_decay, bert_clf_steps, params)

    global_step = 0
    correct_train, count_train = 0, 0
    for seq, mask, tok_type, label in tqdm.tqdm(dataloader, total=bert_clf_steps):
        global_step += 1
        seq = seq.to(device)
        mask = mask.to(device)
        tok_type = tok_type.to(device)
        label = label.to(device)

        outputs = model(seq, mask, tok_type, labels=label)
        loss, logits = outputs[:2]

        count_train += seq.size(0)
        correct_train += (logits.argmax(dim=1).eq(label)
                          .float().sum().detach().cpu().numpy())

        opt.zero_grad()
        loss.backward()
        opt.step()
        sche.step()

        if global_step % bert_clf_period_summary == 0:
            summary.add_scalar("clf_train/loss", loss, global_step)
            summary.add_scalar("clf_train/error_rate", 1 - correct_train / count_train,
                               global_step)
            correct_train, count_train = 0, 0

        if global_step % bert_clf_period_val == 0:
            run_evaluate(model, dataloader_val_iter,
                         bert_clf_val_steps, summary, global_step, device)

        if global_step % bert_clf_period_save == 0 or global_step == bert_clf_steps:
            ckpt_path = os.path.join(model_dir, model_init + "-%04dk" % (global_step // 1000))
            model.save_pretrained(ckpt_path)
            logger.info("BERT classifier saved at %s.", ckpt_path)

        if global_step >= bert_clf_steps:
            break
    model.eval()
    return model


class BertClassifier(ClassifierBase):
    """BERT classifier prediction on paraphrase_list.

    This metric is special, it does not compare the original and paraphrased sentence.
    Instead, it outputs the classifier prediction on paraphrase_list. So we should not compute
    mean or std on this metric.

    Args:
        dataset_name (str): the name of the dataset.
        trainset (dict): a fibber dataset.
        testset (dict): a fibber dataset.
        bert_gpu_id (int): the gpu id for BERT model. Set -1 to use CPU.
        bert_clf_steps (int): steps to train a classifier.
        bert_clf_bs (int): the batch size.
        bert_clf_lr (float): the learning rate.
        bert_clf_optimizer (str): the optimizer name.
        bert_clf_weight_decay (float): the weight decay in the optimizer.
        bert_clf_period_summary (int): the period in steps to write training summary.
        bert_clf_period_val (int): the period in steps to run validation and write validation
            summary.
        bert_clf_period_save (int): the period in steps to save current model.
        bert_clf_val_steps (int): number of batched in each validation.
    """

    def __init__(self, dataset_name, trainset, testset, bert_gpu_id=-1,
                 bert_clf_steps=20000, bert_clf_bs=32, bert_clf_lr=0.00002,
                 bert_clf_optimizer="adamw", bert_clf_weight_decay=0.001,
                 bert_clf_period_summary=100, bert_clf_period_val=500,
                 bert_clf_period_save=5000, bert_clf_val_steps=10, **kargs):

        super(BertClassifier, self).__init__()

        if trainset["cased"]:
            model_init = "bert-base-cased"
            logger.info(
                "Use cased model in BERT classifier prediction metric.")
        else:
            model_init = "bert-base-uncased"
            logger.info(
                "Use uncased model in BERT classifier prediction metric.")

        self._tokenizer = BertTokenizerFast.from_pretrained(
            resources.get_transformers(model_init), do_lower_case="uncased" in model_init)

        if bert_gpu_id == -1:
            logger.warning("BERT metric is running on CPU.")
            self._device = torch.device("cpu")
        else:
            logger.info("BERT metric is running on GPU %d.", bert_gpu_id)
            self._device = torch.device("cuda:%d" % bert_gpu_id)

        self._model = load_or_train_bert_clf(
            model_init=model_init,
            dataset_name=dataset_name,
            trainset=trainset,
            testset=testset,
            bert_clf_steps=bert_clf_steps,
            bert_clf_lr=bert_clf_lr,
            bert_clf_bs=bert_clf_bs,
            bert_clf_optimizer=bert_clf_optimizer,
            bert_clf_weight_decay=bert_clf_weight_decay,
            bert_clf_period_summary=bert_clf_period_summary,
            bert_clf_period_val=bert_clf_period_val,
            bert_clf_period_save=bert_clf_period_save,
            bert_clf_val_steps=bert_clf_val_steps,
            device=self._device)

    def predict_dist_batch(self, origin, paraphrase_list, data_record=None,
                           paraphrase_field="text0"):
        """Predict the log-probability distribution over classes for one batch.

        Args:
            origin (str): the original text.
            paraphrase_list (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.array): a numpy array of size ``(batch_size * num_labels)``.
        """
        if paraphrase_field == "text0":
            batch_input = self._tokenizer(text=paraphrase_list, padding=True, max_length=200,
                                          truncation=True)
        else:
            assert paraphrase_field == "text1"
            batch_input = self._tokenizer(text=[data_record["text0"]] * len(paraphrase_list),
                                          text_pair=paraphrase_list,
                                          padding=True, max_length=200,
                                          truncation=True)
        with torch.no_grad():
            res = F.log_softmax(self._model(
                input_ids=torch.tensor(batch_input["input_ids"]).to(self._device),
                token_type_ids=torch.tensor(batch_input["token_type_ids"]).to(self._device),
                attention_mask=torch.tensor(batch_input["attention_mask"]).to(self._device)
            )[0], dim=1).detach().cpu().numpy()
        return res

    def predict_dist_example(self, origin, paraphrase, data_record=None, paraphrase_field="text0"):
        """Predict the log-probability distribution over classes for one example.

        Args:
            origin (str): the original text.
            paraphrase (list): a set of paraphrase_list.
            data_record (dict): the corresponding data record of original text.
            paraphrase_field (str): the field name to paraphrase.

        Returns:
            (np.array): a numpy array of size ``(num_labels)``.
        """
        return self.predict_dist_batch(origin, [paraphrase], data_record, paraphrase_field)[0]
